#include <cstring>
#include <vector>
using namespace std;

#include "cv.h"
#include "highgui.h"

#include "AVIWriter.h"
#include "ImgFun.h"
#include "IOFun.h"
#include "Project.h"
#include "MathFun.h"
#include "Dynamics.h"
#include "ARMADynamics.h"
#include "ConstVelDynamics.h"
#include "NoDynamics.h"
#include "SSIDDynamics.h"

/**
    Constructor.
 */
Project::Project() {
    START_COLOR = CV_RGB(0, 255, 0);
    PRED_COLOR = CV_RGB(255, 255, 255);
    STEP_COLOR = CV_RGB(255, 255, 0);
    FINAL_COLOR = CV_RGB(255, 0, 0);

    eamInitted = false;

    kt = NULL;
    displayImg = NULL;
    aviFilename = ktFilename = dynFilename = NULL;
    background = COLOR;
    foreground = PIXELS;
    dynamics = NULL;
    nInitStates = 0;
    initted = NULL;
    initStates = NULL;
}

/** 
    Destructor.
 */
Project::~Project() {
	if (kt != NULL) {
		delete kt;
	}
    if (aviFilename != NULL) {
        delete [] aviFilename;
    }
    if (ktFilename != NULL) {
        delete [] ktFilename;
    }
/*    if (kt != NULL) {
        delete kt;
    }*/
    if (dynFilename != NULL) {
        delete [] dynFilename;
    }
    if (dynamics != NULL) {
        delete dynamics;
    }
    if (initStates != NULL) {
        delete [] initStates;
    }
    if (initted != NULL) {
        delete [] initted;
    }
}

/**
    Opens a new AVI file.  
    @param fname filename of AVI file to open
    @param msg if there is an error, an error message is returned here
    @return true if opening is successful, otherwise return 
        false and an error message in msg.
 */
bool Project::openAVI(const char *fname, char **msg) {
    if (aviFilename != NULL) {
        delete [] aviFilename;
    }
    aviFilename = new char[strlen(fname) + 1];
    strcpy(aviFilename, fname);

    avi.open(fname);

    return true;
}

/**
    Opens the specified kinematic tree file.  
    @param fname filename of kinematic model file to open
    @param msg if there is an error, an error message is returned here
    @return true if successful; otherwise return false and 
        an error message in msg.
 */
bool Project::openKT(const char *fname, char **msg) {
    if (ktFilename != NULL) {
        delete [] ktFilename;
    }
    ktFilename = new char[strlen(fname) + 1];
    strcpy(ktFilename, fname);

    if (kt != NULL) {
        delete kt;
    }

    ifstream fin(fname);
    
    try {
        kt = new KinematicTree(fin);
    } catch (TrackerException te) {
        *msg = new char[256];
        sprintf(*msg, "Could not load kinematic model: %s", te.what());
        fin.close();
        return false;
    }

    fin.close();
    return true;
}

/**
    Opens a dynamic model.
    @param fname filename of dynamic model file to open
    @param msg if there is an error, an error message is returned here
    @return true if successful; otherwise return false and 
        an error message in msg.
 */
bool Project::openDynamics(const char *fname, char **msg) {
    if (dynFilename != NULL) {
        delete [] dynFilename;
    }
    dynFilename = new char[strlen(fname) + 1];
    strcpy(dynFilename, fname);

    if (dynamics != NULL) {
        delete dynamics;
        dynamics = NULL;
    }
    
    ifstream fin(fname);
    if (dynamics == NULL) {
        // Read first line, which tells us what sort of dynamic model
        // we have
        int len = 256;
        char *buf = new char[len];
        IOFun::readStringWithComments(fin, buf, len);

        if (strcmp(buf, "ARMA") == 0) {
            dynamics = new ARMADynamics();
        }
        else if (strcmp(buf, "CONST") == 0) {
            dynamics = new ConstVelDynamics();
        }
        else if (strcmp(buf, "SSID") == 0) {
            dynamics = new SSIDDynamics();
        }
        else if (strcmp(buf, "ZERO") == 0) {
            dynamics = new NoDynamics();
        }
        else {
            // unknown type
            dynamics = NULL;
        }

        delete [] buf;
    }
    
    if (dynamics == NULL) {
        // Unknown file type
        fin.close();
        *msg = new char[256];
        sprintf(*msg, "Unknown dynamics filetype.");
        return false;
    }

    dynamics->readFromFile(fin);
    fin.close();


    // Setup the initial states.  Copy as many initial states as
    // have already been set.
    int newNInitStates = dynamics->getNInitStates();
    bool *newInitted = new bool[newNInitStates];
    Vector *newInitStates = new Vector[newNInitStates];

    for (int i = 0; i < newNInitStates; i++) {
        newInitted[i] = false;
    }

    if (initStates != NULL) {
        for (int i = 0; i < newNInitStates && i < nInitStates; i++) {
            newInitStates[i] = initStates[i];
            newInitted[i] = initted[i];
        }
        delete [] initted;
        delete [] initStates;
    }

    nInitStates = newNInitStates;
    initted = newInitted;
    initStates = newInitStates;

    return true;
}

/**
    Sets an initial state for the specified frame. 
    @param state the kinematic model state for the specified frame
    @param frame index of the frame
 */
void Project::initializeState(Vector state, int frame) {
    if (initStates == NULL || frame >= nInitStates) {
        return;
    }
    initStates[frame] = state;
    initted[frame] = true;
}

/**
    Retrieves one of the initial states.
    @return the initial state for the given frame, if one has been set.  
        Otherwise returns NULL.
 */
Vector * Project::getInitState(int frame) {
    if (frame >= nInitStates) {
        return NULL;
    }
    else if (!initted[frame]) {
        return NULL;
    }
    else {
        return &initStates[frame];
    }
}

/**
   If user has initialized all necessary data, prepares for first step
   of tracking and returns true.  Otherwise, returns false and returns
   a message describing the problem in msg.
    @param msg if there is an error, an error message is returned here
 */
bool Project::readyToTrack(char **msg) {
    if (kt == NULL) {
        *msg = new char[256];
        sprintf(*msg, "Need to specify a kinematic model.");
        return false;
    }
    if (!avi.isOpen()) {
        *msg = new char[256];
        sprintf(*msg, "Need to open an AVI file.");
        return false;
    }
    if (!eam.initialized()) {
        *msg = new char[256];
        sprintf(*msg, "Need to initialize the appearance model.");
        return false;
    }
    if (dynamics == NULL) {
        *msg = new char[256];
        sprintf(*msg, "Need to open a dynamic model.");
        return false;
    }
    for (int i = 0; i < nInitStates; i++) {
        if (!initted[i]) {
            *msg = new char[256];
            sprintf(*msg, "This dynamic model requires %d initial states.  All of them have not yet been specified.",
                    nInitStates);
            return false;
        }
    }

    
    // Everything is ok; do initialization
    
    // Initialize history
    history.erase(history.begin(), history.end());
    frameIdx = nInitStates - 1;

    // Initialize dynamic model hidden state
    vector<Vector> ys;
    for (i = 0; i < nInitStates; i++) {
        ys.push_back(fromFullState(initStates[i]));
    }
    dynamics->estimateX0(ys);

    // Initialize kinematic state
    kt->setState(initStates[nInitStates-1]);

    // Make sure output window is closed
    hide();
    
    return true;
}


/**
    Makes the next step in the tracking process: either a prediction,
    the beginning of registration, or the next iteration in
    registration.
    @return true if there are no more frames, false otherwise.
 */
bool Project::nextStep() {
    if (history.size() == 0 || history.back().type == HistoryItem::FINAL) {
        // First step on a new frame
        return startFrame();
    }
    else if (history.back().type == HistoryItem::START) {
        // Make a prediction
        return predict();
    }
    else {
        // Next step on a frame we have already started
        takeStep();
        return false;
    }
}

/**
    Resets everything to the previous tracking step.
    @return true if already at beginning of history, false otherwise
 */
bool Project::prevStep() {
    if (history.size() == 0) {
        // haven't started yet
        hide();
        return true;
    }
    else if (history.size() == 1) {
        // already on first step of first frame
        display();
        return true;
    }

    HistoryItem hi = history.back();
    if (hi.type == HistoryItem::PRED || hi.type == HistoryItem::FINAL) {
        dynamics->back();
    }
    history.pop_back();
    kt->setState(history.back().state);
    frameIdx = history.back().frame;
    display();
    
    return false;
}

/**
    Does all tracking steps for this frame.
    @return false
 */
bool Project::nextFrame() {
    do {
        nextStep();
    } while (history.back().type != HistoryItem::FINAL);

    return false;
}

/**
    Goes backward through tracking steps to the beginning of the
    current frame.
    @return true if already at beginning of history, false otherwise
 */
bool Project::prevFrame() {
    bool retval;

    do {
        retval = prevStep();
    } while (history.back().type != HistoryItem::START);

    return retval;
}

/**
   Pushes the start state for the next frame (if there is one) onto the 
   history and displays it.  
   @return true if there are no more frames, false otherwise.
 */
bool Project::startFrame() {
    Vector s;

    frameIdx++;

    if (frameIdx >= avi.getLength()) {
        return true;
    }

    HistoryItem hi;
    hi.state = kt->getState();
    hi.frame = frameIdx;
    hi.type = HistoryItem::START;
    history.push_back(hi);

    display();

    return false;
}

/** 
   Makes a prediction for the next frame.  Prepares for registration
   on the next frame.
   @return false (this shouldn't be called unless there is at least one
   more frame).
 */
bool Project::predict() {
    HistoryItem hi;
    
    dynamics->predict();
    hi.state = toFullState(dynamics->observe(), history.back().state);
    hi.frame = frameIdx;
    hi.type = HistoryItem::PRED;
    history.push_back(hi);

    kt->setState(hi.state);

    reg.prepForRegistration(avi.getGrayFrame(frameIdx), eam, *kt, curScore);
    
    display();

    return false;
}
    
/**
   Takes the next step in registration on the current frame.  
   @return true if frame registration ends, false otherwise.
  */
bool Project::takeStep() {
    bool finished = reg.oneIteration(eam, *kt, curScore);

    HistoryItem hi;
    hi.state = kt->getState();
    hi.type = finished ? HistoryItem::FINAL : HistoryItem::STEP;
    hi.frame = frameIdx;
    history.push_back(hi);
    
    display();

    if (finished) {
        // Update the dynamics' state and velocity
        dynamics->correct(fromFullState(hi.state));
    }

    return finished;
}

/**
   Tracks all frames of the video and saves the output as a 
   video in the file specified by filename.  Does not display
   any output along the way.
   @param fname name of AVI file to write output to.
 */
void Project::trackAll(const char *fname) {
    // vars having to do with output video
    AVIWriter outputAVI;
    IplImage *frame, *img;
    CvSize frameSize, fontSize;
    CvFont font;
    int baseline;
    Vector s, pred, regfinal, corrected;

    
    // Initialize the font
    cvInitFont(&font, CV_FONT_VECTOR0, 0.5, 0.5, 0.0, 1.0);
    cvGetTextSize("FOOg", &font, &fontSize, &baseline);

    // Initialize some tracking stuff
    frame = avi.getFrame(0);
    frameSize = cvSize(frame->width, frame->height);
    //prevState = initStates[0];

    // Open output video
    outputAVI.open(fname, 29.99, frameSize);


    // Draw the initial states
    for (int i = 0; i < nInitStates; i++) {
        kt->setState(initStates[i]);
        img = avi.getFrame(i);
        kt->drawLines(img, CV_RGB(255, 255, 255));
        cvPutText(img, "init state", cvPoint(20, 20), &font, 
                  CV_RGB(255, 255, 255));
        outputAVI.append(img);
//        prevState = initStates[i];

    }
	
    frameIdx = nInitStates - 1;
	prevState = initStates[frameIdx];

    // Draw the output of state x0 -- note that estimateX0 should have been
    // previously called by readyToTrack().
    s = toFullState(dynamics->observe(), prevState);
    kt->setState(s);
    img = avi.getFrame(frameIdx);
    kt->drawLines(img, CV_RGB(255, 0, 0));
    cvPutText(img, "x0", cvPoint(20, 20), &font, CV_RGB(255, 0, 0));
    outputAVI.append(img);

    
    // Set model to initial (hand-set) state
    kt->setState(initStates[frameIdx]);

    // Tracking loop
    for ( ; frameIdx < avi.getLength(); frameIdx++) {
        // Make a prediction via the LDS model.  This will serve as starting
        // point for the image registration.
		dynamics->predict();
        pred = toFullState(dynamics->observe(), prevState);
        kt->setState(pred);
        
        // Display that prediction
        img = avi.getFrame(frameIdx);
        kt->drawLines(img, CV_RGB(255, 0, 0));
        cvPutText(img, "prediction", cvPoint(20, 20), &font, CV_RGB(255, 0, 0));
        outputAVI.append(img);

        // Perform image registration
        reg.prepForRegistration(avi.getGrayFrame(frameIdx), eam, *kt, curScore);
        while (!reg.oneIteration(eam, *kt, curScore)) {
            img = avi.getFrame(frameIdx);
            kt->drawLines(img, CV_RGB(255, 255, 0));
            outputAVI.append(img);
        }

        // Draw the registration result
        img = avi.getFrame(frameIdx);
        kt->drawLines(img, CV_RGB(0, 255, 0));
        cvPutText(img, "registration result", cvPoint(20, 20), &font, CV_RGB(0, 255, 0));
        outputAVI.append(img);

        // Save the registration result; used for the lengths only later on
        regfinal = kt->getState();
        prevState = regfinal;

        // Correct the LDS hidden state based on this new "observation"
        dynamics->correct(fromFullState(regfinal));

        // Draw the output of the corrected state
        corrected = toFullState(dynamics->observe(), prevState);
        kt->setState(corrected);
        img = avi.getFrame(frameIdx);
        kt->drawLines(img, CV_RGB(0, 0, 255));
        cvPutText(img, "correction", cvPoint(20, 20), &font, CV_RGB(255, 0, 0));
        outputAVI.append(img);

        history.push_back(HistoryItem(frameIdx, corrected, HistoryItem::FINAL));

        // Draw the prediction, final registration, and correction all at once
        img = avi.getFrame(frameIdx);
        kt->setState(pred);
        eam.draw(img, kt, 255, 0, 0);
        baseline = 20;
        cvPutText(img, "prediction", cvPoint(20, baseline), &font, CV_RGB(255, 0, 0));
        kt->setState(regfinal);
        eam.draw(img, kt, 0, 255, 0);
        baseline += MathFun::round(1.1*fontSize.height);
        cvPutText(img, "registration", cvPoint(20, baseline), &font, CV_RGB(0, 255, 0));
        kt->setState(corrected);
        eam.draw(img, kt, 0, 0, 255);
        baseline += MathFun::round(1.1*fontSize.height);
        cvPutText(img, "correction", cvPoint(20, baseline), &font, CV_RGB(0, 0, 255));
        outputAVI.append(img);
    }
    outputAVI.close();

    cvReleaseImage(&frame);
    cvReleaseImage(&img);
}



/**
   Shows the window with the current state superimposed on the current
   frame.  The type of background and foreground information displayed
   is determined by the background and foreground variables.
 */
void Project::display() {
	CvScalar color;
//    IplImage *grayFrame, *edgeImg, *distImg;

    if (history.size() == 0) {
        hide();
        return;
    }

    // This block should be redundant, but just in case...
    frameIdx = history.back().frame;
    if (frameIdx <= 0 || frameIdx >= avi.getLength()) {
        hide();
        return;
    }

    // Prepare background
    frame = avi.getFrame(frameIdx);
    if (displayImg == NULL) {
        displayImg = ImgFun::newImage(cvSize(frame->width, frame->height));
    }

    switch (background) {
    default:
        // fall through
    case COLOR:
        cvCopy(frame, displayImg);
        break;
    case EDGES:
        //grayFrame = avi.getGrayFrame(frameIdx);
        //edgeImg = eam.getEdgeImage(grayFrame);
        //cvConvertImage(edgeImg, displayImg);
        //cvReleaseImage(&grayFrame);
        //cvReleaseImage(&edgeImg);
        break;
    case DISTANCE:
        //distImg = eam.getDistImage();
        //cvCopy(distImg, displayImg);
        //cvReleaseImage(&distImg);
        break;
    }

    // Done with frame now
    cvReleaseImage(&frame);

    // Select foreground drawing color
    switch (history.back().type) {
    case HistoryItem::START:
        color = START_COLOR;
        break;
    case HistoryItem::PRED:
        color = PRED_COLOR;
        break;
    case HistoryItem::STEP:
        color = STEP_COLOR;
        break;
    case HistoryItem::FINAL:
        color = FINAL_COLOR;
        break;
    }

    // Draw foreground
    switch (foreground) {
    case BOXES:
        kt->drawBoxes(displayImg, color);
        break;
    case LINES:
        kt->drawLines(displayImg, color);
        break;
    default:
        // fall through
    case PIXELS:
        eam.draw1(displayImg, kt, color);
        break;
    }

    //eam.drawDerivative(displayImg, kt, CV_RGB(0,0,255));
    //eam.drawJacobian(displayImg, kt, 8, CV_RGB(0,0,255));
    cvNamedWindow(P_WINDOW_NAME, CV_WINDOW_AUTOSIZE);
    cvShowImage(P_WINDOW_NAME, displayImg);
}

/**
    Hides the output window.
 */
void Project::hide() {
    cvDestroyWindow(P_WINDOW_NAME);
}

/**
   Writes the final state estimate for each frame to the specified file.
   @param fout output filestream
 */
void Project::saveHistory(ofstream &fout) {
    Vector s;
    
    vector<HistoryItem>::iterator itr = history.begin();
    for ( ; itr != history.end(); itr++) {
        if (itr->type == HistoryItem::FINAL) {
            s = itr->state;
            for (int i = 0; i < s.len(); i++) {
                fout << s.get(i) << " ";
            }
            fout << endl;
        }
    }
}

/**
   Converts the short state used by the dynamics class (which may not
   include lengths or root position, or might use relative instead
   of absolute root position) to the full state used by the kinematic
   model.
    @param v shortened state vector
    @param extra a full state vector from which the missing parts of
        v are taken
    @return the new full state vector
 */
Vector Project::toFullState(Vector v, Vector extra) {
    //return v;

    // In this case, add root position and lengths
    Vector w(v.len() + 14);
    int j = 0;
    for (int i = 0; i < w.len(); i++) {
        if (i%2 == 1 || i == 2) {
            w.set(i, extra.get(i));
        }
        else {
            w.set(i, v.get(j));
            j++;
        }
    }
    return w;
    
    // In this case, we need to add the root position
    /*
    Vector w(v.len() + 2);
    int j = 0;
    for (int i = 0; i < w.len(); i++) {
        if (i == 1 || i == 2) {
            w.set(i, extra.get(i));
        }
        else {
            w.set(i, v.get(j));
            j++;
        }
    }
    return w;
    */
}


/**
   Converts the full state used by the kinematic model to a different
   state used by the dynamic model.
   @param v full state vector
   @return the shortened state vector
 */
Vector Project::fromFullState(Vector v) {
    return v;

    // In this case, we remove root position as well as lengths
/*    Vector w(v.len() - 14);
    int j = 0;
    for (int i = 0; i < v.len(); i++) {
        if (i%2 != 1 && i != 2) {
            w.set(j, v.get(i));
            j++;
        }
    }
    return w;
*/
    /*
    // In this case, we simply remove the root position
    Vector w(v.len() - 2);
    
    int j = 0 ;
    for (int i = 0; i < v.len(); i++) {
        if (i != 1 && i != 2) {
            w.set(j, v.get(i));
            j++;
        }
    }
    return w;
    */
}


/**
    Writes information pertaining to this project to the specified
    file.
    @param fout output filestream
 */
void Project::writeToFile(ofstream &fout) {
	int i, j;

    fout << "% AVI filename" << endl;
    if (aviFilename == NULL) {
        fout << " " << endl;
    }
    else {
        fout << aviFilename << endl;
    }
    
    fout << "% Dynamics filename" << endl;
    if (dynFilename == NULL) {
        fout << " " << endl;
    }
    else {
        fout << dynFilename << endl;
    }

    //fout << "% Kinematic model filename" << endl;
    //if (ktFilename == NULL) {
    //    fout << " " << endl;
    //}
    //else {
    //    fout << ktFilename << endl;
    //}

    fout << "% Kinematic model" << endl;
    kt->writeToFile(fout);

    fout << "% initial states" << endl;
    fout << nInitStates << endl;
    for (i = 0; i < nInitStates; i++) {
        fout << initStates[i].len() << endl;
        for (j = 0; j < initStates[i].len(); j++) {
            fout << initStates[i].get(j) << " ";
        }
        fout << endl;
    }
    
    fout << "% appearance model" << endl;
    eam.writeToFile(fout);

	reg.writeToFile(fout);
	
}


/**
    Tries to load the project in the specified file.  
    @param fin input filestream
    @param msg if there is an error, an error message is returned here
    @return true on success; on failure, return false and an 
        error message in msg.
 */
bool Project::readFromFile(ifstream &fin, char **msg) {

	int i,j;
    int bufLen = 512;
    char buf[512];
    
    // Get aviFilename
    IOFun::readStringWithComments(fin, buf, bufLen);
    if (strcmp(" ", buf) != 0) {
        if (!openAVI(buf, msg)) {
            return false;
        }
    }

    // Get dynFilename
    IOFun::readStringWithComments(fin, buf, bufLen);
    if (strcmp(" ", buf) != 0) {
        if (!openDynamics(buf, msg)) {
            return false;
        }
    }
    
    // Get kinematic model
    try {
        kt = new KinematicTree(fin);
    } catch (TrackerException te) {
        *msg = new char[256];
        sprintf(*msg, "Could not load kinematic model: %s", te.what());
        return false;
    }

    // Get initial states
    int n = IOFun::readIntWithComments(fin);
    nInitStates = dynamics->getNInitStates();
    initStates = new Vector[nInitStates];
    for (i = 0; i < nInitStates && i < n; i++) {
        int len = IOFun::readIntWithComments(fin);
        initStates[i].initialize(len);
        for (j = 0; j < len; j++) {
            initStates[i].set(j, IOFun::readDoubleWithComments(fin));
        }
        initted[i] = true;
    }

    // Get appearance model
    eam.readFromFile(fin);

	reg.readFromFile(fin);

	fin.close();

    return true;
}

/**
  Display vector in console. For debug
*/
void Project::display(Vector v)
{
	cout << "vector length = " << v.len() << endl;
	for(int i=0; i<v.len(); i++)
		cout << v.get(i) << " ";
	cout << endl;
}