#ifndef PROJECT_INCLUDED
#define PROJECT_INCLUDED

/**************************************************
  Non-GUI project.
 *************************************************/

#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

#include "cv.h"

#include "AVIViewer.h"
#include "Dynamics.h"
#include "EdgeAppearanceModel.h"
#include "PatchAppearanceModel.h"
#include "KinematicTree.h"
#include "Registration.h"
#include "TrackerException.h"
#include "Vector.h"

#define P_WINDOW_NAME ("project window")

/** @defgroup ctrl_group Control Module */
/* @{ */

/** 
    Contains state vector for one step in the tracking process.
    A "step" means the initial state for an image registration,
    a predicted state for the next frame, the end of an iteration
    in image registration, or the final result of an image
    registration.
 */
struct HistoryItem {
    /** The saved state vector */
    Vector state;

    /** The type of step this item corresponds to */
    enum typeType {
        /** The initial state for an image registration */
        START, 
        /** A predicted state for the next frame */
        PRED, 
        /** The end of an iteration in image registration */
        STEP, 
        /** The final result of an image registration */
        FINAL
    } type;
    
    /** The index of the frame this step corresponds to */
    int frame;
    
    /** Constructor */
    HistoryItem() { }

    /** 
        Constructor.
        @param pFrame index of frame
        @param pState state vector
        @param pType type of step
     */
    HistoryItem(int pFrame, Vector pState, typeType pType) {
        frame = pFrame;
        state = pState;
        type = pType;
    }
};


/**
    The main control class for the tracking algorithm.  This contains
    information about the tracking process and code for implementing
    tracking.
 */
class Project {
private:
//public:
    /** History of all steps in the tracking process */
    vector<HistoryItem> history;
    
    /** Current value of the objective function */
    double curScore;

    /** Constant colors used in drawing */
    CvScalar START_COLOR, PRED_COLOR, STEP_COLOR, FINAL_COLOR;

    /** Filenames for AVI, kinematic tree, and dynamic files */
    char *aviFilename, *ktFilename, *dynFilename;

    /** The image where output is drawn for display in a window */
    IplImage *displayImg;
    
    /** The video frame currently being processed */
    IplImage *frame;

    /** The instance of the Registration class being used */
    Registration reg;

    /** Index of the video frame currently being processed */
    int frameIdx;

    
    /** Final state estimate for previous frame */
    Vector prevState;

    /** Number of initial states used */
    int nInitStates;

    /** 
        Array of bools, one per initial state; an element is true
        if the corresponding initial state has been set.
     */
    bool *initted;

    /** The list of initial states; an array of length nInitStates */
    Vector *initStates;
    
    /** True if the appearance model has been initialized */
    bool eamInitted;

    bool startFrame();
    bool predict();
    bool takeStep();

    Vector toFullState(Vector v, Vector extra);
    Vector fromFullState(Vector v);

	void display(Vector v);  // display vector in console. (for debug)

public:
	/** The video we're doing tracking on */
    AVIViewer avi;

    /** The kinematic model being used */
	KinematicTree *kt;

    /** The appearance model */
	PatchAppearanceModel eam;

    /** The dynamic model */
    Dynamics *dynamics;

    /** Type of background to use in output window */
    enum { 
        /** Show color video frame in background */
        COLOR,
        /** Show edge transform of current video frame */
        EDGES, 
        /** Show chamfer distance map of current video frame */
        DISTANCE
    } background;
    
    /** Type of foreground to show in output */
    enum { 
        /** Draw boxes around the locations of the links in the image */
        BOXES, 
        /** Draw lines and circles to represent the model state */    
        LINES, 
        /** Draw the outline of the edges of the appearance model */
        PIXELS 
    } foreground;

	Project();
	~Project();

    void display();
    void hide();

    bool openAVI(const char *fname, char **msg);
    bool openKT(const char *fname, char **msg);
    bool openDynamics(const char *fname, char **msg);
    void initializeState(Vector state, int frame);
    Vector * getInitState(int frame);
    bool readyToTrack(char **msg);

    bool nextStep();
    bool prevStep();

    bool nextFrame();
    bool prevFrame();

    void trackAll(const char *fname);

    void saveHistory(ofstream &fout);

    void writeToFile(ofstream &fout);
    bool readFromFile(ifstream &fin, char **msg);
};
/* @} */

#endif // !def PROJECT_INCLUDED
