#ifndef UPF_INCLUDED
#define UPF_INCLUDED

/**************************************************
  Unscented Particle Filter Tracker.
 *************************************************/
#include "UPFDataType.h"
#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

#include "cv.h"

#include "AVIViewer.h"
#include "AVIWriter.h"

#define UPF_WINDOW_NAME ("Unscented Particle Filter")


/** Some parameters for Unscented Kalman Filter. */
#define ALPHA 1
#define BETA  0
#define KAPPA 2

#define step 100  // to scale chamfer distance

/** @defgroup ctrl_group Control Module */
/* @{ */

/** 
    Contains state vector for one step in the tracking process.
    A "step" means the initial state for an image registration,
    a predicted state for the next frame, the end of an iteration
    in image registration, or the final result of an image
    registration.
 */
struct UPFHistoryItem {
    /** The saved state vector */
    Vector state;

    /** The type of step this item corresponds to */
    enum typeType {
        /** The initial state for an image registration */
        START, 
        /** A predicted state for the next frame */
        PRED, 
        /** The end of an iteration in image registration */
        STEP, 
        /** The final result of an image registration */
        FINAL
    } type;
    
    /** The index of the frame this step corresponds to */
    int frame;

	/** joint centers of the tracker at each frame step. */
	vector<CvPoint> jointCtrs;

	/** tracker score at each frame step. */
	double score;
    
    /** Constructor */
    UPFHistoryItem() { }

    /** 
        Constructor.
        @param pFrame index of frame
        @param pState state vector
        @param pType type of step
     */
    UPFHistoryItem(int pFrame, Vector pState, vector<CvPoint> pjointCtrs, double pScore, typeType pType) {
        frame = pFrame;
        state = pState;
		jointCtrs = pjointCtrs;
		score = pScore;
        type = pType;
    }
};


/**
    The main control class for the tracking algorithm.  This contains
    information about the tracking process and code for implementing
    tracking.
 */
class UPF {

private:
	double Neff, Nthres;

    /** History of all steps in the tracking process */
    vector<UPFHistoryItem> history;

    /** Constant colors used in drawing */
    CvScalar START_COLOR, PRED_COLOR, STEP_COLOR, FINAL_COLOR;

	CvSize frameSize, fontSize;
    CvFont font;

	char text[100];

    /** Filenames for AVI, kinematic tree, and dynamic files */
    char *aviFilename, *ktFilename, *dynFilename;

    /** The video output */
	AVIWriter outputAVI;
	
	/** The image where output is drawn for display in a window */
    IplImage *displayImg;
    
    /** The video frame currently being processed */
    IplImage *frame;

    /** Index of the video frame currently being processed */
    int frameIdx;
    
    /** Final state estimate for previous frame */
    Vector prevState;
	
	Vector pred, regfinal, corrected;
	
	/** Sigma States, Sigma Weights, Prediction Covariance Matrix */
	vector<Vector> sigmaPts;
	vector<double> sigmaWts;
	vector<double> sigmaObs;

	Matrix PxPred;  // covariance of states
	Matrix sigmaX;  // sigmaX * sigmaX = Px
	double Py;  // covariance of observation
	Vector Pxy; // covariance of state and observation
	Vector K;   // a vector, 'cause the observation is 1-d
	
	Vector mx;   // mean of state
	double my;   // mean of observation

	Vector mx_move;  // mean of state, used in sample_move()
	Matrix invP;     // inverse of covariance matrix, used in sample_move()

	Matrix U;     // matrix and vector used in obtaining root of covariance matrix
	Matrix V;
	Matrix D;
	Matrix P;
	Vector tmp;
	Vector xnew;
	Vector xold;
	Matrix mdiff; 
	Matrix *hessian;
	
    /** True if the appearance model has been initialized */
    bool eamInitted;

/*    bool startFrame();
    bool predict();
    bool takeStep();
*/
    Vector toFullState(Vector v, Vector extra);
    Vector fromFullState(Vector v);
    
    int pick_base_sample();
    void calculate_base_weights();
    void predict_sample_position(int new_sample, int old_sample);
	void initialize_sample_weights();
    void set_up_prior_conditions();
    void calculate_shape_parameters_MAP();
    void calculate_shape_parameters();
	void processFrame();
	void registration(int new_sample);
	void drawPred(int new_sample);
	void drawRegistr(int new_sample);
	void drawAllSamples();
	void drawResult();
	void drawCorrection(int new_sample);
	void sample_move(int new_sample, int old_sample);
	void sample_move_registered(int new_sample, int old_sample);

	/** functions related to Unscented Kalman Filter. */
	// scaled unscented transform: generating sigma points and weights
	void scaledUT(int old_sample);
	void propagate(int new_sample, int old_sample);
	void UKF(int new_sample, int old_sample);
	void evaluate_weights(int new_sample, int old_sample);

	

	/** returns a vector of size len, which is from Gaussian. */
	Vector randn(int len);
    
    void display(Vector &v);  // display vector in console. (for debug)
	void showData();   // display all the data in the initialized model. (for debug)
	void display(Matrix &m);
	void vector2matrix(Vector &v, Matrix &m);  // copy v to m, in order to do matrix manipulation

public:
	/** The video we're doing tracking on */
    AVIViewer avi;

    UPFGlobalData m_global;
    UPFIterationData m_iteration;
	
	Vector tracker;

	/** number of acceptance and rejection made in sample_move() per frame */
	vector<int> naccept, nreject;

	/** joint centers of the tracker at each frame step. */
	vector<CvPoint> jointCtrs;

	/** tracker score at each frame step. */
	double trackerScore;

    /** Type of background to use in output window */
   enum { 
        /** Show color video frame in background */
        COLOR,
        /** Show edge transform of current video frame */
        EDGES, 
        /** Show chamfer distance map of current video frame */
        DISTANCE
    } background;
    
    /** Type of foreground to show in output */
    enum { 
        /** Draw boxes around the locations of the links in the image */
        BOXES, 
        /** Draw lines and circles to represent the model state */    
        LINES, 
        /** Draw the outline of the edges of the appearance model */
        PIXELS 
    } foreground;

	UPF();
	~UPF();

    void display();
    void hide();

    bool openAVI(const char *fname, char **msg);
    bool openKT(const char *fname, char **msg);
    bool openDynamics(const char *fname, char **msg);
//    void initializeState(Vector state, int frame);

    bool readyToTrack(char **msg);

//    bool nextStep();
//    bool prevStep();

//    bool nextFrame();
//    bool prevFrame();

    void trackAll(const char *fname);
	void batchrun();

    void saveHistory(ofstream &fout);

	void saveJointCtrs(ofstream &fout);

	void saveTrackerScore(ofstream &fout);

	void saveSampleMoveInfo(ofstream &fout);

    void writeToFile(ofstream &fout);
    bool readFromFile(ifstream &fin, char **msg, int& nsamples);
};
/* @} */

#endif // !def UPF_INCLUDED