#ifndef APPEARANCE_MODEL_INCLUDED
#define APPEARANCE_MODEL_INCLUDED

#include <iostream>
#include <fstream>
using namespace std;

#include "cv.h"
#include "KinematicTree.h"
#include "Matrix.h"
#include "Vector.h"


/** @defgroup app_group Appearance Model */
/* @{ */
/**
   Abstract base class of all appearance models.
 */
class AppearanceModel {
public:
    /**
        Checks to see if the model has been initialized.
        @return true if the model has been initialized
     */
    virtual bool initialized() = 0;
    
    /**
        Initializes the model.
        @param n number of links, number of link images, number of origin poitns
        @param linkImgs array of pointers to IplImages, one per link.  These should
            be single-channel images.  The pixels in these images should have one 
            of three values:
                - 0 -- non-edge pixel
                - 128 -- boundary pixel
                - 255 -- edge pixel
        @param origins coordinates of the bottom left point of the corresponding
                linkImg in the original image.
        @param kt the kinematic model 
    */
    virtual void initialize(int n, IplImage **linkImgs, 
                            CvPoint *origins, KinematicTree *kt) = 0;
        
    /**
        Sets the image that will be used for subsequent calls to getJacobian(),
            getScore(), and getResidual().  
        @param img image to use; should be single-channel.
     */
    virtual void setImage(IplImage *img) = 0;

    /** 
        Gets the transpose of the residual Jacobian, evaluated at the current
        state of kt.
        @param kt kinematic model, with state to evaluate Jacobian at
        @param img image to evaluate Jacobian on.  If NULL uses the image previously
            set by setImage().  If not NULL, calls setImage()
        @return the transpose of the residual Jacobian.
     */        
    virtual Matrix * getJacobianT(KinematicTree *kt, IplImage *img = NULL) = 0;
    
    /**
        Add optional weights on the diagonal of jacobian' * jacobian
        @param JtJ reference to jacobian' * jacobian
     */
    virtual void addJacobianWeights(Matrix &JtJ) = 0;
    
    /**
        Gets the score (residual' * residual) of the model 
        @param kt kinematic model.  Gives state at which to evaluate score.
        @param img image to evaluate Jacobian on.  If NULL uses the image previously
            set by setImage().  If not NULL, calls setImage()
        @return the score
     */
    virtual double getScore(KinematicTree *kt, IplImage *img = NULL) = 0;
    
    /**
        Gets the residual vector
        @param kt kinematic model.  Gives state at which to evaluate residual
        @param img image to evaluate Jacobian on.  If NULL uses the image previously
            set by setImage().  If not NULL, calls setImage()
        @return the residual vector.
     */
    virtual Vector * getResidual(KinematicTree *kt, IplImage *img = NULL) = 0;
    
    /**
        Draws a representation of the appearance model on the given image
        @param img Image to draw on
        @param kt kinematic model, including kinematic state
        @param red red component of color
        @param green green component of color
        @param blue blue component of color
     */
    virtual void draw(IplImage *img, KinematicTree *kt, 
                      int red = 255, int green = 255, int blue = 255) = 0;

    /**
        Alias for draw(); can pass color as a CvScalar instead of 3 ints
        @param img image to draw on
        @param kt kinematic model, including kinematic state
        @param color color to do drawing in
        @see draw()
     */
    virtual void draw1(IplImage *img, KinematicTree *kt, CvScalar color) {
        draw(img, kt, (int)color.val[2], (int)color.val[1], (int)color.val[0]);
    }

    /**
        Alias for draw()
        @see draw() 
     */
    virtual void draw3(IplImage *img, KinematicTree *kt, int red,
                       int green, int blue) {
        draw(img, kt, red, green, blue);
    }

    /** 
        Writes description of the appearance model to a file
        @param fout output filestream
     */
    virtual void writeToFile(ofstream &fout) = 0;
    
    /**
        Reads a description of the model from a file
        @param fin input filestream
     */
    virtual void readFromFile(ifstream &fin) = 0;
};
/* @} */

#endif // !def APPEARANCE_MODEL_INCLUDED
