#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
using namespace std;

#include "cv.h"
#include "highgui.h"

#include "EdgeAppearanceModel.h"
#include "ImgFun.h"
#include "IOFun.h"
#include "KinematicTree.h"
#include "Matrix.h"
#include "ModelPoint.h"
#include "MathFun.h"
#include "Vector.h"

EdgeAppearanceModel::EdgeAppearanceModel() { 
    distMap = NULL;
    lowThresh  = LOW_THRESH;
    highThresh = HIGH_THRESH;
    edgePts = NULL;
    nLinks = 0;
    didInit = false;
}

EdgeAppearanceModel::~EdgeAppearanceModel() {
    if (edgePts != NULL) {
        delete [] edgePts;
    }
    if (distMap != NULL) {
        cvReleaseMat(&distMap);
    }
}


/*
   Return true iff model has been initialized.
 */
bool EdgeAppearanceModel::initialized() {
    return didInit;
}

/*
   Collect the edge pixels for each link.
   n -- number of links, number of link images, number of origin poitns
   linkImgs -- array of pointers to IplImages, one per link.  These should
     be single-channel images.  The pixels in these images should have one 
     of three values:
       0 -- non-edge pixel
       128 -- boundary pixel
       255 -- edge pixel
   origins -- coordinates of the bottom left point of the corresponding
     linkImg in the original image.
   kt -- the kinematic model we are using
 */
void EdgeAppearanceModel::initialize(int n, IplImage **linkImgs, 
                                     CvPoint *origins, KinematicTree *kt) {
    const SPMLink *link;
    unsigned char *data;
    int step;
    CvSize size;
    ModelPoint mp;

    nLinks = n;
    nPts = 0;
    if (edgePts != NULL) {
        delete edgePts;
    }
    edgePts = new vector<ModelPoint>[n];

    for (int i = 0; i < n; i++) {
        link = (*kt)[i];
        cvGetRawData(linkImgs[i], (uchar**)&data, &step, &size);
        step /= sizeof(data[0]);
        
        // Loop over every pixel in the link image
        for (int x = 0; x < size.width; x++) {
            for (int y = 0; y < size.height; y++) {
                if (data[y*step + x] == 255) {
                    // Convert to link coordinate system to make bounds
                    // checking easy.
                    mp = link->worldToLocal(ModelPoint(x + origins[i].x,
                                                       y + origins[i].y));
                    
                    // Check to be sure the point is inside the bounds of
                    // the link
                    if (link->contains(mp)) {
                        edgePts[i].push_back(mp);
                    }
                }
            }
        }
        nPts += edgePts[i].size();
    }
    didInit = true;
}

/**
   Sets the thresholds used in the Canny edge detector 
   @param low low threshold
   @param high high threshold
 */
void EdgeAppearanceModel::setEdgeThresholds(int low, int high) {
    lowThresh  = low;
    highThresh = high;
}

/**
   Retrieves the thresholds used in the Canny edge detector
   @param low low threshold
   @param high high threshold
 */
void EdgeAppearanceModel::getEdgeThresholds(int &low, int &high) {
    low  = lowThresh;
    high = highThresh;
}

/*
   Set the image that will be used for subsequent calls to getJacobian(),
   getScore(), and getResidual().  distMap is calculated, and it stays
   the same until setImage() is called again or one of the 3 functions 
   is called with img != NULL.
   
   img should be a single-channel image.
 */
void EdgeAppearanceModel::setImage(IplImage *img) {
    int w = img->width;
    int h = img->height;
    CvSize size;
    size.width = w;
    size.height = h;

    // Set up the distance map matrix
    if (distMap != NULL) {
        cvReleaseMat(&distMap);
    }
    distMap = cvCreateMat(h, w, CV_32FC1);

    // Get the edges
    IplImage *iplEdges = getEdgeImage(img);

    // Flip the edge output so it is the format cvDistTransform() expects
    //iplNot(iplEdges, iplEdges);
    cvNot(iplEdges, iplEdges);

    // Run the distance transform
    cvDistTransform(iplEdges, distMap, CV_DIST_L2, CV_DIST_MASK_5, NULL);
    
    // All of this is to find the maximum distance (maxDist); the other values
    // get thrown out.
    double minVal;
    CvPoint minLoc, maxLoc;
    cvMinMaxLoc(distMap, &minVal, &maxDist, &minLoc, &maxLoc);

    // Optionally scale the distance map
    //cvConvertScale(distMap, distMap, 100);


    cvReleaseImage(&iplEdges);
}

/*
   Get the transpose of the jacobian, evaluated for the current state of kt.
 */
Matrix * EdgeAppearanceModel::getJacobianT(KinematicTree *kt, IplImage *img) {
    if (img != NULL) {
        setImage(img);
    }

    int ncols = kt->getStateLen();
    Matrix *jac = new Matrix(ncols, nPts);
    Matrix *partJac = new Matrix(2, ncols);
    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    double dDdx, dDdy;
    int row = 0;

    jac->setZero();

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = edgePts[i].begin();
        for (; itr != edgePts[i].end(); itr++, row++) {
            mp = link->localToWorld(*itr);
            if (inBounds(mp.x, mp.y)) {
                derivative(mp.x, mp.y, dDdx, dDdy);
            }
            else {
                // Not sure what the right thing to do is, but let's
                // try this
                dDdx = 0;
                dDdy = 0;
            }
            link->getKinematicJacobian(partJac, mp, NULL);
            for (int col = 0; col < ncols; col++) {
                jac->set(col, row, dDdx*partJac->get(0,col) 
                                   + dDdy*partJac->get(1,col));
            }
        }
    }

	delete partJac;

    return jac;
}

/**

    Calculate Hessian = J'*J

 */

Matrix * EdgeAppearanceModel::getHessian(KinematicTree *kt, IplImage *img) {

    Matrix *Jt = getJacobianT(kt, img);

    Matrix *H = new Matrix(Jt->rows(), Jt->rows());

    MathFun::timesTranspose(*Jt, *H);

	delete Jt;

    return H;

}

/* 
   Add optional weights on the diagonal of the Jacobian transpose
   * Jacobian.
 */
void EdgeAppearanceModel::addJacobianWeights(Matrix &JtJ) {
    for (int i = 0; i < JtJ.rows(); i++) {
        if (i == 0 || ((i >=4) && (i%2 == 0))) {
            JtJ.set(i, i, JtJ.get(i,i) + 1000);
        }
        else {
            JtJ.set(i, i, JtJ.get(i,i) + .001);
        }
    }
}

double EdgeAppearanceModel::getScore(KinematicTree *kt, IplImage *img) {
    Vector *r = getResidual(kt, img);
    double score = (*r)*(*r);
    delete r;
    return score; // score == dot product of residual vector with itself
}

Vector * EdgeAppearanceModel::getResidual(KinematicTree *kt, IplImage *img) {
    if (img != NULL) {
        setImage(img);
    }

    Vector *r = new Vector(nPts);
    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    int vidx = 0;
    int x, y;
    ModelPoint mp;

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = edgePts[i].begin();
        for ( ; itr != edgePts[i].end(); itr++, vidx++) {
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            if (inBounds(x, y)) {
                r->set(vidx, cvmGet(distMap, y, x));
            }
            else {
                r->set(vidx, 2*maxDist);
            }
        }
    }

    return r;
}

/**
    Returns an edge image of img.
    @param img pointer to original image.  Should be a single-channel image.
    @return a pointer to the edge image.  When finished, calling code should
        release this image by calling cvReleaseImage()
 */
IplImage * EdgeAppearanceModel::getEdgeImage(IplImage *img) {
    CvSize size = cvSize(img->width, img->height);
    
    //IplImage *iplEdges = ImgFun::newImage(size, IPL_DEPTH_8U, 1);
    IplImage *iplEdges = cvCreateImage(size, IPL_DEPTH_8U, 1);

    // Run the edge detector
    cvCanny(img, iplEdges, lowThresh, highThresh, APERTURE_SIZE);

    return iplEdges;
}

/**
    Gets the chamfer distance map.
   @return The distance map represented as a color IplImage.  If
   the input image has not been set yet, returns NULL.
 */
IplImage * EdgeAppearanceModel::getDistImage() {
    if (distMap == NULL) {
        return NULL;
    }

    // Scale the distance map to a grayscale image
    IplImage *bw = ImgFun::newImage(cvSize(distMap->width, distMap->height),
                                    IPL_DEPTH_8U, 1);
    cvConvertScale(distMap, bw, 20);

    // Convert that grayscale to a color image
    IplImage *img = ImgFun::newImage(cvSize(bw->width, bw->height),
                                     IPL_DEPTH_8U, 3);
    cvConvertImage(bw, img);
    cvReleaseImage(&bw);
    
    return img;
}

/*
   Draw the edge points on an image.
 */
void EdgeAppearanceModel::draw(IplImage *img, KinematicTree *kt,
                                    int red, int green, int blue) {
    char color[3];
    color[0] = blue;
    color[1] = green;
    color[2] = red;

    int w = img->width;
    int h = img->height;
    int x, y;

    int s = 0;

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = edgePts[i].begin();
        for ( ; itr != edgePts[i].end(); itr++) {
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            if (x >= 0 && x < w && y >= 0 && y < h) {
                //iplPutPixel(img, x, h - y - 1, color);
                cvRectangle(img, cvPoint(x-s,y-s), cvPoint(x+s,y+s),
                            CV_RGB(red, green, blue), CV_FILLED);
            }
        }
    }
}

/**
   Draws the derivative field
   @param img the image to draw on
   @param kt the kinematic model
   @param color color to use to draw derivative
 */
void EdgeAppearanceModel::drawDerivative(IplImage *img, KinematicTree *kt,
                                         CvScalar color) {
    int w = img->width;
    int h = img->height;
    int x, y, x2, y2;
    double dDdx, dDdy;

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = edgePts[i].begin();
        for (int j = 0; itr != edgePts[i].end(); j++, itr++) {
            if (j % 8 != 0) {
                continue;
            }
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            derivative(x, y, dDdx, dDdy);
            x2 = MathFun::round(mp.x - 8*dDdx);
            y2 = MathFun::round(mp.y - 8*dDdy);
            if (x >= 0 && x < w && y >= 0 && y < h
                    && x2 >= 0 && x2 < w && y2 >= 0 && y2 < h) {
                cvLine(img, cvPoint(x,y), cvPoint(x2,y2), color);
            }
        }
    }
}

/**
    Draws a representation of one column of the residual Jacobian field
    @param img image to draw on
    @param kt the kinematic model
    @param stateIdx the index of the state to draw the Jacobian for
    @param color color to draw in
 */
void EdgeAppearanceModel::drawJacobian(IplImage *img, KinematicTree *kt, 
                                       int stateIdx, CvScalar color) {
    int w = img->width;
    int h = img->height;
    int x, y, x2, y2;

    int ncols = kt->getStateLen();
    Matrix *partJac = new Matrix(2, ncols);
    Matrix *jac = getJacobianT(kt);

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    int j = 0;

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = edgePts[i].begin();
        for ( ; itr != edgePts[i].end(); j++, itr++) {
            if (j % 8 != 0) {
                continue;
            }
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            //link->getKinematicJacobian(partJac, mp, NULL);
            //s = partJac->get(0, stateIdx) * partJac->get(0,stateIdx)
            //         + partJac->get(1, stateIdx) * partJac->get(1, stateIdx);
            //n = (s == 0 ? 1 : sqrt(s));
            //x2 = MathFun::round(mp.x + partJac->get(0,stateIdx));
            //y2 = MathFun::round(mp.y + partJac->get(1,stateIdx));
            
            //x2 = MathFun::round(mp.x + jac->get(stateIdx, j));
            //y2 = MathFun::round(mp.y);
            
            if (jac->get(stateIdx, j) > 0) {
                x2 = x + 10;
            }
            else if (jac->get(stateIdx, j) == 0) {
                x2 = x;
            }
            else {
                x2 = x - 10;
            }
            y2 = y;
            
            if (x >= 0 && x < w && y >= 0 && y < h
                    && x2 >= 0 && x2 < w && y2 >= 0 && y2 < h) {
                cvLine(img, cvPoint(x,y), cvPoint(x2,y2), color);
            }
        }
    }

    delete partJac;
    delete jac;
}


/*
   Write edgePts to a file
 */
void EdgeAppearanceModel::writeToFile(ofstream &fout) {
    vector<ModelPoint>::iterator itr;

    fout << lowThresh << endl;
    fout << highThresh << endl;
    fout << nLinks << endl;
    for (int i = 0; i < nLinks; i++) {
        fout << edgePts[i].size() << endl;
        itr = edgePts[i].begin();
        for ( ; itr != edgePts[i].end(); itr++) {
            fout << itr->x << " " << itr->y << endl;
        }
    }
}

/*
   Read edgePts from a file.  Note that everything that is initialized
   in initialize() should also be set by calling this.
 */
void EdgeAppearanceModel::readFromFile(ifstream &fin) {
    int ptsInLink;
    double x, y;

    lowThresh = IOFun::readIntWithComments(fin);
	highThresh = IOFun::readIntWithComments(fin);

    nLinks = IOFun::readIntWithComments(fin);
	nPts = 0;
    if (edgePts != NULL) {
        delete edgePts;
    }
    edgePts = new vector<ModelPoint>[nLinks];
    
    for (int i = 0; i < nLinks; i++) {
        ptsInLink = IOFun::readIntWithComments(fin);
		nPts += ptsInLink;
        for (int j = 0; j < ptsInLink; j++) {
            x = IOFun::readDoubleWithComments(fin);
			y = IOFun::readDoubleWithComments(fin);
			edgePts[i].push_back(ModelPoint(x, y));
        }
    }
    didInit = true;
}

/**
   Calculate the derivative of distMap at the given x,y.
   Assumes the point (x,y) is in bounds, so you should
   check that before calling this.
   @param x x-coordinate
   @param y y-coordinate
   @param dDdx reference in which to return x component of derivative
   @param dDdy reference in which to return y component of derivative
 */
void EdgeAppearanceModel::derivative(int x, int y, 
                                     double &dDdx, double &dDdy) {
    if (x == 0) {
        dDdx = cvmGet(distMap, y, x + 1) - cvmGet(distMap, y, x);
    }
    else if (x == distMap->cols - 1) {
        dDdx = cvmGet(distMap, y, x) - cvmGet(distMap, y, x - 1);
    }
    else {
        dDdx = (cvmGet(distMap, y, x + 1) - cvmGet(distMap, y, x - 1)) / 2;
    }

    if (y == 0) {
        dDdy = cvmGet(distMap, y + 1, x) - cvmGet(distMap, y, x);
    }
    else if (y == distMap->rows - 1) {
        dDdy = cvmGet(distMap, y, x) - cvmGet(distMap, y - 1, x);
    }
    else {
        dDdy = (cvmGet(distMap, y + 1, x) - cvmGet(distMap, y - 1, x)) / 2;
    }
}

/**
   Returns true if x,y is a point in the current distance map.
 */
bool EdgeAppearanceModel::inBounds(int x, int y) {
    assert(distMap != NULL);

    if (x < 0 || x >= distMap->cols 
            || y < 0 || y >= distMap->rows) {
        return false;
    }
    else {
        return true;
    }
}
