#ifndef EDGE_APPEARANCE_MODEL_INCLUDED
#define EDGE_APPEARANCE_MODEL_INCLUDED

#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

#include "cv.h"
#include "AppearanceModel.h"
#include "KinematicTree.h"
#include "Matrix.h"
#include "ModelPoint.h"
#include "Vector.h"


/** @addtogroup app_group */
/* @{ */
/** 
    An appearance model that uses edge features.
 */
class EdgeAppearanceModel : public AppearanceModel {
private:
    /**
        Default values for Canny edge detector parameters.
     */
    enum {
        /** Default low threshold for Canny edge detector */
        LOW_THRESH = 300, 
        /** Default high threshold for Canny edge detector */
        HIGH_THRESH = 900, 
        /** Default aperture size for Canny edge detector */
        APERTURE_SIZE = 5
    };

    /** Chamfer distance map */
    CvMat *distMap;

    /** number of links in the model */
    int nLinks;
    
    /** number of edge pixels in the model */
    int nPts;

    /** list of edge pixels */
    vector<ModelPoint> *edgePts;

    /** The largest distance in the distance map */
    double maxDist;                 

    /** Edge thresholds used by Canny edge detector */
    int lowThresh, highThresh;

    /** True if model has been initialized */
    bool didInit;

    void derivative(int x, int y, double &dDdx, double &dDdy);
    bool inBounds(int x, int y);

public:
    EdgeAppearanceModel();
    ~EdgeAppearanceModel();
    
    bool initialized();
    void initialize(int n, IplImage **linkImgs, 
                    CvPoint *origins, KinematicTree *kt);
    
    void setEdgeThresholds(int low, int high);
    void getEdgeThresholds(int &low, int &high);
    void setImage(IplImage *img);
    Matrix * getJacobianT(KinematicTree *kt, IplImage *img = NULL);
	Matrix * getHessian(KinematicTree *kt, IplImage *img = NULL);
    void addJacobianWeights(Matrix &JtJ);
    double getScore(KinematicTree *kt, IplImage *img = NULL);
    Vector * getResidual(KinematicTree *kt, IplImage *img = NULL);
    
    IplImage *getEdgeImage(IplImage *img);
    IplImage *getDistImage();

    void draw(IplImage *img, KinematicTree *kt, int red = 255, int green = 255,
              int blue = 255);
    void drawDerivative(IplImage *img, KinematicTree *kt, CvScalar color);
    void drawJacobian(IplImage *img, KinematicTree *kt, int stateIdx, CvScalar color);

    void writeToFile(ofstream &fout);
    void readFromFile(ifstream &fin);
};
/* @} */

#endif // !def EDGE_APPEARANCE_MODEL_INCLUDED
