#include <iostream>
#include <fstream>
using namespace std;

#include "cv.h"
#include "Vector.h"
#include "Matrix.h"
#include "KinematicTree.h"
#include "SPMLink.h"

/**
    Constructor; builds tree given input file with model description.
    @param fin file describing kinematic model
 */
KinematicTree::KinematicTree(ifstream &fin) : root(fin) {
    nStates = root.countStates();
    root.listLinks(links);
}

/**
	Copy Constructor
 */


/**
    Sets model state
    @param v state vector
    @return true if successful
 */
bool KinematicTree::setState(const Vector &v) {
	savedState = v;
	int i = 0;
	return (root.setState(v, i));
}

/**
    Gets pointer to a link in the model
    @param i index of link to get
    @return pointer to SPMLink object associated with link i
 */
const SPMLink * KinematicTree::operator[](const int i) const {
    return links[i];
}

/**
    @return length of state vector
 */
int KinematicTree::getStateLen() {
	return nStates;
}

/**
    @return number of links in model
 */
int KinematicTree::countLinks() {
    return links.size();
}

/**
    @return current state of kinematic model
 */
Vector KinematicTree::getState() {
    Vector state(nStates);
    int idx = 0;
    root.getState(state, idx);
    return state;
}

/**
    Gets the locations of points used to control the model state with
        the mouse
    @param controlPts reference to a vector where the control points will be
        inserted
 */
void KinematicTree::getControlPts(vector<CvPoint> &controlPts) {
    root.getControlPts(controlPts);
}

/**
    Gets the joint centers in the kinematic model.
    @param jointCtrs references to a vector where the joint centers will be inserted
 */
void KinematicTree::getJointCenters(vector<CvPoint> &jointCtrs) {

    root.getJointCenters(jointCtrs);

}

/**
    Given the new locations of control points, sets the kinematic state
    @param controlPt location of the point that moved
    @param cidx the index of the changed control point
 */
void KinematicTree::setStateFromControlPts(CvPoint controlPt, int cidx) {
    int i = 0;
    root.setStateFromControlPts(controlPt, i, cidx);
}

/**
    Save model to file.  Current state is used as initial state.
    @param fout output filestream
 */
void KinematicTree::writeToFile(ofstream &fout) {
    root.writeToFile(fout);
}

/**
    Draw the kinematic model as a set of filled rectangles.
    @param image image to draw on
    @param color color to draw with
 */
void KinematicTree::drawFilled(IplImage *image, CvScalar color) {
	root.drawFilled(image, color);
}

/**
    Draw the kinematic model as a set of rectangular outlines
    @param image image to draw on
    @param color color to draw with
 */
void KinematicTree::drawBoxes(IplImage *image, CvScalar color) {
	root.drawBoxes(image, color);
}

/**
    Draw the kinematic model as a set of lines
    @param image image to draw on
    @param color color to draw with
 */
void KinematicTree::drawLines(IplImage *image, CvScalar color) {
	root.drawLines(image, color);
}

/**
    Draw the kinematic model as a set of thick lines and largish
    circles
    @param image image to draw on
    @param color color to draw with
    @param r radius of balls to use
 */
void KinematicTree::drawSticksAndBalls(IplImage *image, CvScalar color, 
                                       int r) {
    root.drawSticksAndBalls(image, color, r);
}
