#include "Matrix.h"
#include "Vector.h"
#include "cv.h"

#define NULL 0

Matrix::Matrix() {
    matrix = NULL;
}

Matrix::Matrix(int rows, int cols) {
    matrix = NULL;
    initialize(rows, cols);
}

Matrix::Matrix(const Matrix &m) {
    matrix = NULL;
    duplicate(m);
}

Matrix::~Matrix() {
    destroy();
}

/**
    Sets the size of the matrix and allocates memory
    @param rows number of rows
    @param cols number of columns
    @param type OpenCV type of the data in the matrix
 */
void Matrix::initialize(int rows, int cols) {
    assert(rows > 0 && cols > 0);
    destroy();
    matrix = cvCreateMat(rows, cols, CV_64FC1);
}

void Matrix::destroy() {
    if (matrix != NULL) {
        cvReleaseMat(&matrix);
    }
    matrix = NULL;
}


/*
    Accessors
 */
int Matrix::width() const {
    if (matrix == NULL) {
        return 0;
    }
    else {
        return matrix->cols;
    }
}

int Matrix::height() const {
    if (matrix == NULL) {
        return 0;
    }
    else {
        return matrix->rows;
    }
}

/**
    Gets element r,c of the matrix
    @param r zero-based row index
    @param c zero-based column index
    @return value at (r,c) in the matrix
 */
double Matrix::get(int r, int c) const {
    assert(matrix != NULL);
    
    return cvmGet(matrix, r, c);
}

/**
    Sets element r,c of the matrix
    @param r zero-based row index
    @param c zero-based column index
    @param d new value of (r,c) in the matrix
 */
void Matrix::set(int r, int c, double d) {
    assert(matrix != NULL);

    cvmSet(matrix, r, c, d);
}

/**
    Sets every element of the matrix to zero.
 */
void Matrix::setZero() {
    assert(matrix != NULL);

    cvSetZero(matrix);
}

/**
	Sets the matrix an Identity
	The matrix must be square
 */
void Matrix::setIdentity() {
	assert(matrix != NULL);
	
	cvSetIdentity(matrix);
}


/*
   Other
 */
Vector Matrix::operator*(const Vector &v) const {
    assert(width() == v.len());

    Vector retval(height());

    cvMatMulAdd(matrix, v.vector, NULL, retval.vector);

    return retval;
}

Matrix Matrix::operator*(double dbl) const {
	assert(matrix != NULL);
	Matrix retval(rows(), cols());

	for (int i = 0; i < rows(); i++)
		for (int j = 0; j < cols(); j++)
		{
			retval.set(i,j, this->get(i,j) * dbl);
		}
	
	return retval;
}

Matrix Matrix::operator*(const Matrix &m) {
    assert(cols() == m.rows());

    Matrix retval(rows(), m.cols());

    cvMatMulAdd(matrix, m.matrix, NULL, retval.matrix);

    return retval;
}

Matrix Matrix::operator+(const Matrix &m) {
    assert(cols() == m.cols() && rows() == m.rows());

    Matrix retval(rows(), cols());

    cvScaleAdd(matrix, cvRealScalar(1.0), m.matrix, retval.matrix);

    return retval;
}

Matrix Matrix::operator-(const Matrix &m) {
    assert(cols() == m.cols() && rows() == m.rows());

    Matrix retval(rows(), cols());

    cvScaleAdd(m.matrix, cvRealScalar(-1.0), matrix, retval.matrix);

    return retval;
}

Matrix Matrix::operator=(const Matrix &m) {
    duplicate(m);

    return *this;
}

/**
    @return inverse of matrix
 */
Matrix Matrix::inv() {
    int method = 0;
    Matrix retval;

    if (rows() == cols()) {
        double det = cvDet(matrix);
        if (det != 0) {
            method = CV_LU;
        }
        else {
            method = CV_SVD;
        }
        retval.initialize(rows(), cols());
    }
    else {
        method = CV_SVD;
        retval.initialize(cols(), rows());
    }

    cvInvert(matrix, retval.matrix, method);
    
    return retval;
}

double Matrix::det()
{
	return cvDet(matrix);
}

void Matrix::SVD(Matrix &U, Matrix &V, Matrix &D)
{
	assert(U.matrix != NULL && V.matrix != NULL && D.matrix != NULL);

	cvSVD(matrix, D.matrix, U.matrix, V.matrix, CV_SVD_V_T);
}

/*void Matrix::display()
{
	for (int i=0; i < rows(); i++)
		for (int j=0; j < cols(); j++)
			if (j != cols() - 1)
				cout << get(i,j) << " " ;
			else
				cout << get(i,j) << endl;
}
*/

Matrix Matrix::transpose() {
    Matrix retval(cols(), rows());
    cvTranspose(matrix, retval.matrix);
    return retval;
}

void Matrix::duplicate(const Matrix &m) {
    destroy();

    if (m.matrix == NULL) {
        matrix = NULL;
    }
    else {
        matrix = cvCloneMat(m.matrix);
    }
}
