#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
using namespace std;

#include "cv.h"
#include "highgui.h"

#include "PatchAppearanceModel.h"
#include "ImgFun.h"
#include "IOFun.h"
#include "KinematicTree.h"
#include "Matrix.h"
#include "ModelPoint.h"
#include "MathFun.h"
#include "Vector.h"

PatchAppearanceModel::PatchAppearanceModel() { 
    frameImg = NULL;
    dxImg = dyImg = NULL;
    imgData = NULL;
    patchPts = NULL;
}

PatchAppearanceModel::~PatchAppearanceModel() {
    if (frameImg != NULL) {
        cvReleaseImage(&frameImg);
    }
    if (dxImg != NULL) {
        cvReleaseImage(&dxImg);
    }
    if (dyImg != NULL) {
        cvReleaseImage(&dyImg);
    }
    if (patchPts != NULL) {
        delete [] patchPts;
    }
}

/*
   Return true iff model has been initialized.
 */
bool PatchAppearanceModel::initialized() {
    return didInit;
}

/*
   Collect the edge pixels for each link.
   n -- number of links, number of link images, number of origin poitns
   linkImgs -- array of pointers to IplImages, one per link.  These should
     be single-channel (grayscale) images.  
   origins -- coordinates of the bottom left point of the corresponding
     linkImg in the original image.
   kt -- the kinematic model we are using
 */
void PatchAppearanceModel::initialize(int n, IplImage **linkImgs, 
                                      CvPoint *origins, KinematicTree *kt) {
    const SPMLink *link;
    unsigned char *data;
    int step;
    CvSize size;
    ModelPoint mp;

    nLinks = n;
    nPts = 0;
    patchPts = new vector<ModelPoint>[n];

    int ctr = 0;

    for (int i = 0; i < n; i++) {
        link = (*kt)[i];
        cvGetRawData(linkImgs[i], (uchar**)&data, &step, &size);
        step /= sizeof(data[0]);
        
        // Loop over every pixel in the link image
        for (int x = 0; x < size.width; x++) {
            for (int y = 0; y < size.height; y++) {
                // Convert to link coordinate system to make bounds
                // checking easy.
                mp = link->worldToLocal(ModelPoint(x + origins[i].x,
                                                   y + origins[i].y));
                mp.value = data[y*step + x];
                // Check to be sure the point is inside the bounds of
                // the link
                if (link->contains(mp)) {
                    //if (ctr % STRIDE == 0) {
                        patchPts[i].push_back(mp);
                    //}
                    //ctr++;
                }
            }
        }
        nPts += patchPts[i].size();
    }
}

/*
   Set the image that will be used for subsequent calls to getJacobian(),
   getScore(), and getResidual().  It stays
   the same until setImage() is called again or one of the 3 functions 
   is called with img != NULL.
   
   img should be a single-channel image.
 */
void PatchAppearanceModel::setImage(IplImage *img) {
    CvSize size;
    
    // Make a copy of the image
    imgSize = cvSize(img->width, img->height);
    if (frameImg == NULL) {
        //frameImg = ImgFun::newImage(imgSize, IPL_DEPTH_8U, 1);
        frameImg = ImgFun::newImage(imgSize, IPL_DEPTH_32F, 1);
    }
    else if (imgSize.width != frameImg->width 
                || imgSize.height != frameImg->height) {
        cvReleaseImage(&frameImg);
        //frameImg = ImgFun::newImage(imgSize, IPL_DEPTH_8U, 1);
        frameImg = ImgFun::newImage(imgSize, IPL_DEPTH_32F, 1);
    }
    //cvCopy(img, frameImg);
    cvConvertScale(img, frameImg, 1.0);
    cvGetRawData(frameImg, (uchar**)&imgData, &imgDataStep, &size);
    imgDataStep /= sizeof(imgData[0]);   

    // Calculate derivatives
    if (dxImg != NULL) {
        cvReleaseImage(&dxImg);
    }
    if (dyImg != NULL) {
        cvReleaseImage(&dyImg);
    }
    dxImg = ImgFun::newImage(imgSize, IPL_DEPTH_32F, 1);
    dyImg = ImgFun::newImage(imgSize, IPL_DEPTH_32F, 1);
    cvSobel(frameImg, dxImg, 1, 0); //, CV_SCHARR);
    frameImg->origin = IPL_ORIGIN_TL;
    cvSobel(frameImg, dyImg, 0, 1); //, CV_SCHARR);
    frameImg->origin = IPL_ORIGIN_BL;

    cvConvertScale(dxImg, dxImg, 1.0/8.0);
    cvConvertScale(dyImg, dyImg, 1.0/8.0);

    cvGetRawData(dxImg, (uchar**)&dxData, &dxDataStep, &size);
    dxDataStep /= sizeof(dxData[0]);
    cvGetRawData(dyImg, (uchar**)&dyData, &dyDataStep, &size);
    dyDataStep /= sizeof(dyData[0]);


    //cvNamedWindow("dx", CV_WINDOW_AUTOSIZE);
    //cvShowImage("dx", dxImg);
    //cvNamedWindow("dy", CV_WINDOW_AUTOSIZE);
    //cvShowImage("dy", dyImg);

}

/*
   Get the transpose of the jacobian, evaluated for the current state of kt.
 */
Matrix * PatchAppearanceModel::getJacobianT(KinematicTree *kt, IplImage *img) {
    if (img != NULL) {
        setImage(img);
    }

    int ncols = kt->getStateLen();
    Matrix *jac = new Matrix(ncols, nPts);
    Matrix *partJac = new Matrix(2, ncols);
    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    double dDdx, dDdy, dDdx2, dDdy2;
    int row = 0;

    jac->setZero();

    //ofstream fout("jac.txt");
    //ofstream fout2("der.txt");

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = patchPts[i].begin();
        for (; itr != patchPts[i].end(); itr++, row++) {
            mp = link->localToWorld(*itr);
            if (inBounds(MathFun::round(mp.x), MathFun::round(mp.y))) {
                dDdx = dxData[MathFun::round(mp.y)*dxDataStep 
                              + MathFun::round(mp.x)];
                dDdy = dyData[MathFun::round(mp.y)*dyDataStep 
                              + MathFun::round(mp.x)];
                //derivative(MathFun::round(mp.x), MathFun::round(mp.y),
                //           mp.value, dDdx2, dDdy2);
            }
            else {
                // Not sure what the right thing to do is, but let's
                // try this
                dDdx = 0;
                dDdy = 0;
            }
            link->getKinematicJacobian(partJac, mp, NULL);
            for (int col = 0; col < ncols; col++) {
                jac->set(col, row, dDdx*partJac->get(0,col) 
                                   + dDdy*partJac->get(1,col));
                //fout << jac->get(col, row) << " ";
                //fout2 << dDdx << " " << partJac->get(0,col) << " " << dDdy << " " << partJac->get(1,col) << " / ";
            }
            //fout << endl;
            //fout2 << endl;
        }
    }

    //fout.close();
    //fout2.close();

    delete partJac;

    return jac;
}

/**
    Calculate Hessian = J'*J
 */
Matrix * PatchAppearanceModel::getHessian(KinematicTree *kt, IplImage *img) {
    Matrix *Jt = getJacobianT(kt, img);
    Matrix *H = new Matrix(Jt->rows(), Jt->rows());
    MathFun::timesTranspose(*Jt, *H);
    delete Jt;
    return H;
}


/* 
   Add optional weights on the diagonal of the Jacobian transpose
   * Jacobian.
 */
void PatchAppearanceModel::addJacobianWeights(Matrix &JtJ) {
    for (int i = 0; i < JtJ.rows(); i++) {
        JtJ.set(i, i, JtJ.get(i,i) + .001);
    }
}

double PatchAppearanceModel::getScore(KinematicTree *kt, IplImage *img) {
    Vector *r = getResidual(kt, img);
    double score = (*r)*(*r);
    delete r;
    return score/nPts; // score == dot product of residual vector with itself
}

Vector * PatchAppearanceModel::getResidual(KinematicTree *kt, IplImage *img) {
    if (img != NULL) {
        setImage(img);
    }

    Vector *r = new Vector(nPts);
    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    int vidx = 0;
    int x, y;
    ModelPoint mp;
    //double s;

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = patchPts[i].begin();
        for ( ; itr != patchPts[i].end(); itr++, vidx++) {
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            if (inBounds(x, y)) {
                r->set(vidx, imgData[y*imgDataStep + x] - itr->value);
                //r->set(vidx, abs(imgData[y*imgDataStep + x] - itr->value));
                //s = imgData[y*imgDataStep + x] - itr->value;
                //r->set(vidx, s*s);
            }
            else {
                r->set(vidx, 2*255);
            }
        }
    }

    return r;
}

/*
   Draw the appearance model on top of an image.  Draw rectangles around the
   links in the indicated color.
 */
void PatchAppearanceModel::draw(IplImage *img, KinematicTree *kt,
                                int red, int green, int blue) {
    int w = img->width;
    int h = img->height;
    int x, y;
    int val;

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = patchPts[i].begin();
        for ( ; itr != patchPts[i].end(); itr++) {
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            if (x >= 0 && x < w && y >= 0 && y < h) {
                val = itr->value;
                //iplPutPixel(img, x, h - y - 1, &val);
                cvRectangle(img, cvPoint(x,y), cvPoint(x,y),
                            CV_RGB(val, val, val), CV_FILLED);
            }
        }
    }

    kt->drawBoxes(img, CV_RGB(red, green, blue));

}

/*
   Draw the derivative field
 */
void PatchAppearanceModel::drawDerivative(IplImage *img, KinematicTree *kt,
                                          CvScalar color) {
    int w = img->width;
    int h = img->height;
    double x, y;
    int x2, y2;
    double dDdx, dDdy;
    double lx1, lx2, ly1, ly2, lx, ly, l;

    if (frameImg == NULL) {
        IplImage *grayImg = ImgFun::newImage(cvSize(w,h), IPL_DEPTH_8U, 1);
        cvConvertImage(img, grayImg);
        setImage(grayImg);
        cvReleaseImage(&grayImg);
    }

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = patchPts[i].begin();
        for (int j = 0; itr != patchPts[i].end(); j++, itr++) {
            if (j % 5 != 0) {
                continue;
            }
            mp = link->localToWorld(*itr);
            x = mp.x;
            y = mp.y;
            if (!inBounds(MathFun::round(x), MathFun::round(y))) {
                continue;
            }
            //derivative(MathFun::round(x), MathFun::round(y), mp.value, 
            //           dDdx, dDdy);
            dDdx = dxData[MathFun::round(mp.y)*dxDataStep 
                          + MathFun::round(mp.x)];
            dDdy = dyData[MathFun::round(mp.y)*dyDataStep 
                          + MathFun::round(mp.x)];

            // Scale derivative for better visibility
            dDdx *= 100;
            dDdy *= 100;

            // If necessary, scale vector so it stays inside the image
            lx1 = x / dDdx;
            lx2 = (x - (w-1)) / dDdx;
            ly1 = y / dDdy;
            ly2 = (y - (h-1)) / dDdy;
            lx = max(lx1, lx2);
            ly = max(ly1, ly2);
            l = min(lx, ly);
            l = min(l, 1.0);
                
            x2 = MathFun::round(x - l*dDdx);
            y2 = MathFun::round(y - l*dDdy);
            x2 = min(x2, w-1);
            x2 = max(x2, 0);
            y2 = min(y2, h-1);
            y2 = max(y2, 0);
            cvLine(img, cvPoint(MathFun::round(x),MathFun::round(y)),
                   cvPoint(x2, y2), color);
        }
    }
}

void PatchAppearanceModel::drawJacobian(IplImage *img, KinematicTree *kt, 
                                       int stateIdx, CvScalar color) {
    int w = img->width;
    int h = img->height;
    int x, y, x2, y2;

    if (frameImg == NULL) {
        IplImage *grayImg = ImgFun::newImage(cvSize(w,h), IPL_DEPTH_8U, 1);
        cvConvertImage(img, grayImg);
        setImage(grayImg);
        cvReleaseImage(&grayImg);
    }


    int ncols = kt->getStateLen();
    Matrix *partJac = new Matrix(2, ncols);
    Matrix *jac = getJacobianT(kt);

    const SPMLink *link;
    vector<ModelPoint>::iterator itr;
    ModelPoint mp;
    int j = 0;

    for (int i = 0; i < nLinks; i++) {
        link = (*kt)[i];
        itr = patchPts[i].begin();
        for ( ; itr != patchPts[i].end(); j++, itr++) {
            if (j % 5 != 0) {
                continue;
            }
            mp = link->localToWorld(*itr);
            x = MathFun::round(mp.x);
            y = MathFun::round(mp.y);
            //link->getKinematicJacobian(partJac, mp, NULL);
            //s = partJac->get(0, stateIdx) * partJac->get(0,stateIdx)
            //         + partJac->get(1, stateIdx) * partJac->get(1, stateIdx);
            //n = (s == 0 ? 1 : sqrt(s));
            //x2 = MathFun::round(mp.x + partJac->get(0,stateIdx));
            //y2 = MathFun::round(mp.y + partJac->get(1,stateIdx));
            
            //x2 = MathFun::round(mp.x + jac->get(stateIdx, j));
            //y2 = MathFun::round(mp.y);
            
            if (jac->get(stateIdx, j) > 0) {
                x2 = x + 10;
            }
            else if (jac->get(stateIdx, j) == 0) {
                x2 = x;
            }
            else {
                x2 = x - 10;
            }
            y2 = y;
            
            if (x >= 0 && x < w && y >= 0 && y < h
                    && x2 >= 0 && x2 < w && y2 >= 0 && y2 < h) {
                cvLine(img, cvPoint(x,y), cvPoint(x2,y2), color);
            }
        }
    }
    delete partJac;
    delete jac;
}



/*
   Write patchPts to a file
 */
void PatchAppearanceModel::writeToFile(ofstream &fout) {
    vector<ModelPoint>::iterator itr;

    fout << nLinks << endl;
    for (int i = 0; i < nLinks; i++) {
        fout << patchPts[i].size() << endl;
        itr = patchPts[i].begin();
        for ( ; itr != patchPts[i].end(); itr++) {
            fout << itr->x << " " << itr->y << " " << itr->value << endl;
        }
    }
}

/*
   Read patchPts from a file.  Note that everything that is initialized
   in initialize() should also be set by calling this.
 */
void PatchAppearanceModel::readFromFile(ifstream &fin) {
    int ptsInLink;
    double x, y;
    int value;
    int ctr = 0;

    nLinks = IOFun::readIntWithComments(fin);
	nPts = 0;
    patchPts = new vector<ModelPoint>[nLinks];
    
    for (int i = 0; i < nLinks; i++) {
        ptsInLink = IOFun::readIntWithComments(fin);
        for (int j = 0; j < ptsInLink; j++) {
            x = IOFun::readDoubleWithComments(fin);
			y = IOFun::readDoubleWithComments(fin);
            value = IOFun::readIntWithComments(fin);
            if (ctr % STRIDE == 0) {
                patchPts[i].push_back(ModelPoint(x, y, value));
            }
            ctr++;
        }
        nPts += patchPts[i].size();
    }
}

/*
   Calculate the derivative of the image at the given x,y.
   Assumes the point (x,y) is in bounds, so you should
   check that before calling this.
 */
/*
void PatchAppearanceModel::derivative(int x, int y, int val,
                                      double &dDdx, double &dDdy) {
    if (x == 0) {
		//double Ix1 = abs(imgData[y*imgDataStep + x+1] - val);
		//double Ix  = abs(imgData[y*imgDataStep + x] - val);
        //dDdx = Ix1*Ix1 - Ix*Ix;
		//dDdx = Ix1 - Ix;
        dDdx = imgData[y*imgDataStep + x+1] - imgData[y*imgDataStep + x];
    }
    else if (x == imgSize.width - 1) {
		//double Ix = abs(imgData[y*imgDataStep + x] - val);
		//double Ix1 = abs(imgData[y*imgDataStep + x-1] - val);
        //dDdx = Ix*Ix - Ix1*Ix1;
        //dDdx = Ix - Ix1;
		dDdx = imgData[y*imgDataStep + x] - imgData[y*imgDataStep + x-1];
    }
    else {
		//double Ixp1 = abs(imgData[y*imgDataStep + x+1] - val);
		//double Ixm1 = abs(imgData[y*imgDataStep + x-1] - val);
        //dDdx = (Ixp1*Ixp1 - Ixm1*Ixm1) / 2.0;
        //dDdx = (Ixp1 - Ixm1) / 2.0;
		dDdx = (imgData[y*imgDataStep + x+1] 
                - imgData[y*imgDataStep + x-1]) / 2.0;
    }

    if (y == 0) {
        //double Iyp1 = abs(imgData[(y+1)*imgDataStep + x] - val);
		//double Iy   = abs(imgData[y*imgDataStep + x] - val);
        //dDdy = Iyp1*Iyp1 - Iy*Iy;
		//dDdy = Iyp1 - Iy;
        dDdy = imgData[(y+1)*imgDataStep + x] - imgData[y*imgDataStep + x];
    }
    else if (y == imgSize.height - 1) {
		//double Iy   = abs(imgData[y*imgDataStep + x] - val);
		//double Iym1 = abs(imgData[(y-1)*imgDataStep + x] - val);
        //dDdy = Iy*Iy - Iym1*Iym1;
		//dDdy = Iy - Iym1;
        dDdy = imgData[y*imgDataStep + x] - imgData[(y-1)*imgDataStep + x];
    }
    else {
		//double Iyp1 = abs(imgData[(y+1)*imgDataStep + x] - val);
		//double Iym1 = abs(imgData[(y-1)*imgDataStep + x] - val);
        //dDdy = (Iyp1*Iyp1 - Iym1*Iym1) / 2.0;
		//dDdy = (Iyp1 - Iym1) / 2.0;
        dDdy = (imgData[(y+1)*imgDataStep + x] 
                - imgData[(y-1)*imgDataStep + x]) / 2.0;
    }
}
*/

/*
   Returns true if x,y is a point in the current input image.
 */
bool PatchAppearanceModel::inBounds(int x, int y) {
    assert(imgData != NULL);

    if (x < 0 || x >= imgSize.width 
            || y < 0 || y >= imgSize.height) {
        return false;
    }
    else {
        return true;
    }
}
