#include <cmath>



#include <cstdio>



#include "AppearanceModel.h"

#include "IOFun.h"

#include "KinematicTree.h"

#include "Registration.h"

#include "MathFun.h"

#include "Vector.h"



/**

    Constructor.  Initializes TREL and TABS, the termination

    tolerances.

 */

Registration::Registration() {

    // Initialize various search constants

    TREL = .05;

    TABS = .05;



    lowerBounds = NULL;

    upperBounds = NULL;

    enforceBounds = NULL;

}



/**

   Does all iterations of registration for an image.  After execution, 

        the state of kt will be set to the location of the new minimum.

   @param image the image to perform registration on; should be a 

        single-channel image.

   @param am appearance model

   @param kt kinematic model, including initial state.  After execution,

        state of model will be set to location of new minimum.

   @param score the final score of the model is returned here

 */

void Registration::registerImage(IplImage *image, AppearanceModel &am,

        KinematicTree &kt, double &score) {

    prepForRegistration(image, am, kt, score);

    while (!oneIteration(am, kt, score)) ;

}



/**

   Prepares for image registration.  Should be called before oneIteration().

   @param image the image to perform registration on; should be a 

        single-channel image. 

   @param am appearance model

   @param kt kinematic model, including initial state

   @param score the final score of the model is returned here

*/

void Registration::prepForRegistration(IplImage *image, AppearanceModel &am,

        KinematicTree &kt, double &score) {

    

    // Set the size of Vectors used in registration

    int n = kt.getStateLen();

    grad.initialize(n);

    descentDir.initialize(n);

    stateNew.initialize(n);



    // Get ready for registration and calculate the first descent direction

    am.setImage(image);

    getDescentDir(am, kt, descentDir, grad, curScore);

    normg = normg0 = sqrt(grad*grad);

    cnt = 0;



    score = curScore;

}





/**

   Does one iteration of an image registration.  Assumes prepForRegistration()

   has already been called.  After execution, the state of kt will be set

   to the location of the new minimum.



   @param am appearance model

   @param kt kinematic model, including current state.  After execution,

        state of model will be set to location of new minimum.

   @param score the final score of the model is returned here

   @return true if registration terminated with 

        this iteration, false otherwise.

 */

bool Registration::oneIteration(AppearanceModel &am, KinematicTree &kt, 

                                double &score) {

    double newScore;

    Vector stateCur;

    bool lsSuccess; //linesearch success?



    stateCur = kt.getState();

    lsSuccess = goldenSearch(am, kt, descentDir, curScore, newScore);

    //lsSuccess = linesearch(am, kt, descentDir, curScore, newScore);

    //kt.setState(stateCur + descentDir);

    //lsSuccess = true;

    //newScore = am.getScore(&kt);

    stateNew = kt.getState();



    /* Enforce bound constraints, if they exist */

/*    if (upperBounds != NULL && lowerBounds != NULL) {

        for (int i = 0; i < upperBounds->len(); i++) {

            if (enforceBounds[i]) {

                if (stateNew.get(i) < lowerBounds->get(i)) {

                    stateNew.set(i, lowerBounds->get(i));

                }

                else if (stateNew.get(i) > upperBounds->get(i)) {

                    stateNew.set(i, upperBounds->get(i));

                }

            }

        }

        kt.setState(stateNew);

        curScore = am.getScore(&kt);

    }    
*/  // modified by Ping
	if (upperBounds != NULL && lowerBounds != NULL) {

        for (int i = 0; i < upperBounds->len(); i++) {

            if (enforceBounds[i]) {

                if (i%2 == 0 && i != 2) {
					// angle range
					if (myfmod(stateNew.get(i), 6.2832) > upperBounds->get(i))
						stateNew.set(i, upperBounds->get(i));
					else if (myfmod(stateNew.get(i), 6.2832) < lowerBounds->get(i))
						stateNew.set(i, lowerBounds->get(i));
				}
				else if (i%2 == 1 && i != 3) {
					// length range
					if (stateNew.get(i) < lowerBounds->get(i)) {
	                    stateNew.set(i, lowerBounds->get(i));
	                }
					else if (stateNew.get(i) > upperBounds->get(i)) {
						stateNew.set(i, upperBounds->get(i));
					}
				}

            }

        }

        kt.setState(stateNew);

        curScore = am.getScore(&kt);

    }  
	// end of Ping modification

    if (!lsSuccess || stateCur == stateNew) {

        score = curScore;

        return true;

    }

        

    if (cnt < MAX_ITERS && normg > TREL*normg0 + TABS) {

        // Not done yet

        getDescentDir(am, kt, descentDir, grad, curScore);

        normg = sqrt(grad*grad);

        cnt++;

        score = curScore;

        return false;

    }

    else {

        // Done

        score = curScore = am.getScore(&kt);

        return true;

    }

}



/**

    Calculates descent direction.

   @param am appearance model

   @param kt kinematic model, including current state

   @param descentDir the calculated descent direction is returned here

   @param JtR the gradient of the objective function at the current

        state is returned here

   @param score the final score of the model is returned here

 */

void Registration::getDescentDir(AppearanceModel &am, KinematicTree &kt,

        Vector &descentDir, Vector &JtR, double &score) {

    Matrix *Jt;

    Vector *R;

    Matrix JtJ;



    Jt = am.getJacobianT(&kt);

    R  = am.getResidual(&kt);

//   score = (*R)*(*R);
	score = (*R)*(*R) / (double)R->len(); //am.getScore(&kt);  // Ping modified



    // Calculate JtJ

    JtJ.initialize(Jt->rows(), Jt->rows());

    MathFun::timesTranspose(*Jt, JtJ);

    am.addJacobianWeights(JtJ);



    // Calculate JtR

    JtR = (*Jt) * (*R);

    //JtR.initialize(Jt->rows());

    //MathFun::matVecMul((*Jt), *R, JtR);



    // If all components of the gradient (JtR) are exactly zero,

    // JtJ will be singular and hence we won't be able to calculate

    // a descent direction.  So just use the gradient.

    if (JtR.isZero()) {

        descentDir = -JtR;

    }

    else {

        cvSolve(JtJ.matrix, JtR.vector, descentDir.vector, CV_LU);

        descentDir.negate();

    }



    delete Jt;

    delete R;

}





/**

   Conducts a search in the specified direction for a minimum.  

   The search is carried out by starting with a full step in that

   direction, then reducing the stepsize by a factor of 'cutback'

   until a sufficient decrease in the function value is achieved.

 

   Returns true if decrease was achieved, false if it is unable to

   find a sufficiently lower function value.  If the search is successful,

   the state of the model will be set to the new minimum.  Otherwise the

   state of the model will be unchanged.



    @param am appearance model

    @param kt kinematic model, including current state

    @param descentDir a direction of decrease for the objective function

    @param curf the value of the objective function at the current state

    @param newScore the value of the objective function at the new state

        is returned here

    @return true if decrease was achieved, false if unable to find a

        sufficiently lower function value.

    

 */

bool Registration::linesearch(AppearanceModel &am, KinematicTree &kt,

                              Vector &descentDir, double curf, 

                              double &newScore) {

    int lsMaxIters = 10;    // Maximum number of cutbacks

    int cnt = 0;

    double alpha = 1e-6;

    double cutback = .5;

    double step = 1.0;

    double newf;

    Vector initState = kt.getState();



    do {

        kt.setState(initState + descentDir*step);

        newf = am.getScore(&kt);

        step *= cutback;

        cnt++;

    } while (((newf - curf) > 0) && cnt < lsMaxIters);



    if ((newf - curf) <= 0) {

        // Success! -- found an improved function value

        newScore = newf;

        return true;

    }

    else {

        // Failure :( -- didn't improve the function value

        newScore = curf;

        kt.setState(initState);

        return false;

    }

}





/*

   Conducts a golden section search for a minimum in the direction descentDir.

   Return true if a smaller function value is discovered, false otherwise.

   If successful, the state of model will be updated.



    @param am appearance model

    @param kt kinematic model, including current state

    @param descentDir a direction of decrease for the objective function

    @param curf the value of the objective function at the current state

    @param newScore the value of the objective function at the new state

        is returned here

    @return true if decrease was achieved, false if unable to find a

        sufficiently lower function value.

 */

bool Registration::goldenSearch(AppearanceModel &am, KinematicTree &kt, 

        Vector &descentDir, double initf, double &newScore) {

    // Search constants

    double GOLD = 1 - (3-sqrt(5.0))/2;

    double MINB = 1e-3;

    double MAXC = 10;

    double EPS = 1e-4;

    double TOL = 2*sqrt(EPS);

    int maxIters = 10;



    double newf;

    Vector initState = kt.getState();

    double a, b, c;     // The bracketing triplet

    double fa, fb, fc;  // Function values at a, b, c

    double d;

    int cnt = 0;



    // Find an initial bracketing so that a < b < c and fb < fa and fb < fc.

    a = 0;

    fa = initf;

    b = 1.0;

    kt.setState(initState + descentDir*b);

    fb = am.getScore(&kt);



    if (fa <= fb) {

        // Shrink b until fb < fa

        while (fa <= fb) {

            c = b;

            fc = fb;

            b = b*GOLD;

            if (b < MINB) {

                // It looks like we're not going to find a smaller value

                // -- let's give up

                newScore = initf;

                kt.setState(initState);

                return false;

            }

            kt.setState(initState + descentDir*b);

            fb = am.getScore(&kt);

        }

    }

    else {

        // Expand the interval until we find a c with fc > fb

        c = a + (b-a)*(GOLD+1);

        kt.setState(initState + descentDir*c);

        fc = am.getScore(&kt);

        while (fc <= fb && c < MAXC) {    // The c < MAXC part is a hack to keep 

                                          // things from getting out of control.

            if (fc < fb) {

                b = c;

                fb = fc;

            }

            c = a + (c-a)*(GOLD+1);

            kt.setState(initState + descentDir*c);

            fc = am.getScore(&kt);

        }

        if (c >= MAXC) {

            // Couldn't find a bracketing; give up but since fb represents

            // an improvement, set the new state to b

            newScore = fb;

            kt.setState(initState + descentDir*b);

            return true;

        }

    }





    // Now the searching begins:

    while ((c-a) > 2*TOL*b && (c-a) > EPS && cnt < maxIters) {

        if (b-a > c-b) {

            // Choose x between b and a

            d = b - GOLD*(b-a);

            kt.setState(initState + descentDir*d);

            newf = am.getScore(&kt);

            if (newf < fb) {

                c = b;

                fc = fb;

                b = d;

                fb = newf;

            }

            else {

                a = d;

                fa = newf;

            }

        }

        else {

            // Choose x between b and c

            d = b + GOLD*(c-b);

            kt.setState(initState + descentDir*d);

            newf = am.getScore(&kt);

            if (newf < fb) {

                a = b;

                fa = fb;

                b = d;

                fb = newf;

            }

            else {

                c = d;

                fc = newf;

            }

        }



        cnt++;

    }



    newScore = fb;

    return true;

}



/** 

    Write bound constraints information to project file

 */

void Registration::writeToFile(ofstream &fout) {

    fout << "% bound constraints" << endl;
	int i;



    if (upperBounds == NULL || lowerBounds == NULL || enforceBounds == NULL) {

        // No bound constraints

        fout << "0" << endl;

        return;

    }



    fout << "1" << endl;

    int len = lowerBounds->len();

    fout << len << endl;

    for (i = 0; i < len; i++) {

        fout << lowerBounds->get(i) << " ";

    }

    fout << endl;

    for ( i = 0; i < len; i++) {

        fout << upperBounds->get(i) << " ";

    }

    fout << endl;

    for (i = 0; i < len; i++) {

        if (enforceBounds[i]) {

            fout << "1 ";

        }

        else {

            fout << "0 ";

        }

    }

    fout << endl;

}



/**

    Read bound constraints information from the project file.

 */

void Registration::readFromFile(ifstream &fin) {

    int c = IOFun::readIntWithComments(fin);
	int i;

    if (c) {

        int len = IOFun::readIntWithComments(fin);

        if (lowerBounds != NULL) {

            delete lowerBounds;

        }

        lowerBounds = new Vector(len);

        for (i = 0; i < len; i++) {

            lowerBounds->set(i, IOFun::readDoubleWithComments(fin));

        }

        if (upperBounds != NULL) {

            delete upperBounds;

        }

        upperBounds = new Vector(len);

        for (i = 0; i < len; i++) {

            upperBounds->set(i, IOFun::readDoubleWithComments(fin));

        }

        if (enforceBounds != NULL) {

            delete [] enforceBounds;

        }

        enforceBounds = new bool[len];

        int b;

        for (i = 0; i < len; i++) {

            b = IOFun::readIntWithComments(fin);

            enforceBounds[i] = (b != 0);

        }

    }

    else {

        upperBounds = NULL;

        lowerBounds = NULL;

        enforceBounds = NULL;

    }

}