#include <vector>
#include <iostream>
#include <fstream>
#include <cmath>
using namespace std;

#include "cv.h"
#include "Vector.h"
#include "Matrix.h"
#include "ModelPoint.h"
#include "SPMLink.h"
#include "TrackerException.h"
#include "MathFun.h"
#include "IOFun.h"
#include "ImgFun.h"

/**
   This is the root constructor -- it assumes you are creating the
   root link.
   @param fin input filestream.  Current position in stream should be
        beginning of the kinematic model's description.
 */
SPMLink::SPMLink(ifstream &fin) {
    int firstIdx = 0;
    this->parent = NULL;
    name = NULL;
    readFromFile(fin, firstIdx, NULL);
}

/**
    Constructor
    @param parent this link's parent in the kinematic tree
    @param fin input filestream.  Current position in stream should be
        beginning of this link's description.
    @param firstIdx first index where this link's state appears in the model
        state vector
    @param name name of this link
 */
SPMLink::SPMLink(SPMLink *parent, ifstream &fin, int &firstIdx, char *pname) {
    this->parent = parent;
    name = NULL;
    readFromFile(fin, firstIdx, pname);
}

/**
    Destructor
 */
SPMLink::~SPMLink() {
	delete [] name;

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		delete (*itr);
	}
}

/**
   Reads the structure of this link from the specified file, then recursively
   reads the structure of its descendant links.
    @param fin input filestream.  Current position in stream should be
        beginning of this link's description.
    @param firstIdx first index where this link's state appears in the model
        state vector
    @param name name of this link
 */
void SPMLink::readFromFile(ifstream &fin, int &firstIdx, char *pname) {
    char token[MAX_LINE_LEN];

    if (name != NULL) {
        delete [] name;
    }

    if (pname == NULL) {
        IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
        if (token[0] == '\0') {
            throw TrackerException("Expected name, got end of file!");
        }
        name = new char[strlen(token)+1];
        strcpy(name, token);
    }
	else {
		name = new char[strlen(pname)+1];
		strcpy(name, pname);
	}

	// Next token should be '{'
	IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
	if (token[0] != '{') {
		throw TrackerException("Bad file format.");
	}

    // Get binary attributes....
    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': isStateAngle = false; break;
        case '1': isStateAngle = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': isStateLen = false; break;
        case '1': isStateLen = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': isStateThickness = false; break;
        case '1': isStateThickness = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': isStatePos = false; break;
        case '1': isStatePos = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': canSetAngle = false; break;
        case '1': canSetAngle = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': canSetLen = false; break;
        case '1': canSetLen = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': canSetThickness = false; break;
        case '1': canSetThickness = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    switch (token[0]) {
        case '0': canSetPos = false; break;
        case '1': canSetPos = true;  break;
        default:  throw TrackerException("End of file!"); break;
    }


    // Get default configuration
    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    if (token[0] == '\0') {
        throw TrackerException("End of file!");
    }
    localAngle = strtod(token, NULL);

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    if (token[0] == '\0') {
        throw TrackerException("End of file!");
    }
    len = strtod(token, NULL);

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    if (token[0] == '\0') {
        throw TrackerException("End of file!");
    }
    thickness = strtod(token, NULL);

    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    if (token[0] == '\0') {
        throw TrackerException("End of file!");
    }
    position.x = strtod(token, NULL);
    
    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    if (token[0] == '\0') {
        throw TrackerException("End of file!");
    }
    position.y = strtod(token, NULL);

	calcEndPts(false);

    // Figure the indices of this link's first and last positions in the
    // state vector.
    int n = 0;
    if (isStateAngle) { n++; }
    if (isStateLen) { n++; }
    if (isStateThickness) { n++; }
    if (isStatePos) { n += 2; }
    if (n == 0) {
        firstStateIdx = lastStateIdx = -1;
    }
    else {
        firstStateIdx = firstIdx;
        lastStateIdx = firstIdx + n - 1;
        firstIdx = firstIdx + n;
    }

    // Now get any sublinks
    IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
    while (strcmp(token, "}")) {
        if (token[0] == '\0') {
            throw TrackerException("Expected name or }, got end of file!");
        }
        children.push_back(new SPMLink(this, fin, firstIdx, token));
		IOFun::readTokenWithComments(fin, token, MAX_LINE_LEN);
	}
}

/**
    Writes this link's description to a file.
    @param fout output filestream
 */
void SPMLink::writeToFile(ofstream &fout) {
    fout << name << " {" << endl;
    
    fout << (isStateAngle ? "1" : "0") << endl;
    fout << (isStateLen ? "1" : "0") << endl;
    fout << (isStateThickness ? "1" : "0") << endl;
    fout << (isStatePos ? "1" : "0") << endl;
    
    fout << (canSetAngle ? "1" : "0") << endl;
    fout << (canSetLen ? "1" : "0") << endl;
    fout << (canSetThickness ? "1" : "0") << endl;
    fout << (canSetPos ? "1" : "0") << endl;

    fout << localAngle << endl;
    fout << len << endl;
    fout << thickness << endl;
    fout << position.x << endl;
    fout << position.y << endl;

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->writeToFile(fout);
	}

    fout << "}" << endl;
}

/**
   Counts the number of elements in the state vector for this link
        and its descendants.
   @return number of elements in the state description of this link
        and its descendants in the tree
 */
int SPMLink::countStates() {
    int nStates = 0;

    if (isStateAngle)     nStates++;
    if (isStateLen)       nStates++;
    if (isStateThickness) nStates++;
    if (isStatePos)       nStates += 2;

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		nStates += (*itr)->countStates();
	}

	return nStates;
}

/**
   Adds to v an ordered list of pointers to this link and all its descendants.
   @param v vector pointers to SPMLinks 
 */
void SPMLink::listLinks(vector<SPMLink *> &v) {
    v.push_back(this);

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
        (*itr)->listLinks(v);
	}
}

/**
   Sets the state of this link, then recursively set the states of its
   descendants.
   @param v kinematic state vector
   @param idx in v of this link's first state element
 */
bool SPMLink::setState(const Vector &v, int &idx) {
	// Angle
	if (isStateAngle) {
        localAngle = v.get(idx++);
	}

	// Length
	if (isStateLen) {
		len = v.get(idx++);
	}

	// Thickness
	if (isStateThickness) {
		thickness = v.get(idx++);
	}

	// Position
	if (isStatePos) {
		position.x = v.get(idx++);
		position.y = v.get(idx++);
	}

	calcEndPts(false);

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->setState(v, idx);
	}

	return true;
}

/**
   Given this link's state (localAngle, length, thickness, position), 
        calculates its origin and endpoint and axis and world angle.
   @param recurse if true, indicates that the function should be recursively 
        called for descendants of this link in the tree.
 */
void SPMLink::calcEndPts(bool recurse) {
	// Origin
	if (parent == NULL) {
		sx = position.x;
		sy = position.y;
	}
	else {
        ModelPoint mp = parent->localToWorld(position);
		sx = mp.x;
		sy = mp.y;
	}

	// Angle, in world frame
	if (parent != NULL) {
		angle = parent->angle + localAngle;
	}
	else {
		angle = localAngle;
	}

	// Axis
	ax = cos(angle);
    ay = sin(angle);

	// Endpoint
    ex = sx + len*ax;
    ey = sy + len*ay;

	if (recurse) {
		vector<SPMLink*>::iterator itr = children.begin();
		for (; itr != children.end(); itr++) {
			(*itr)->calcEndPts(recurse);
		}
	}
}

/**
   Inserts the state information for this link, and then recursively calls
   this link's descendants to get their state information.
   @param v state vector for kinematic model
   @param idx index of beginning of this link's state information
        in the state vector
 */
void SPMLink::getState(Vector &v, int &idx) {
	if (isStateAngle) {
		v.set(idx++, localAngle);
	}
	if (isStateLen) {
		v.set(idx++, len);
	}
	if (isStateThickness) {
		v.set(idx++, thickness);
	}
	if (isStatePos) {
		v.set(idx++, position.x);
		v.set(idx++, position.y);
	}

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->getState(v, idx);
	}
}


/**
    Evaluates the kinematic part of the jacobian at the given point.

    @param jac pointer to a 2 x nStates matrix.  First row is partials of x
        with respect to the states, 2nd row is partials of y w.r.t. states
    @param mpWorld point, in world coordinates, where we want to evaluate the
        jacobian
    @param child pointer to the SPMLink that called this function.  NULL indicates
        that mpWorld is a point in this link.
 */
void SPMLink::getKinematicJacobian(Matrix *jac, ModelPoint mpWorld, 
                                   const SPMLink *child) const {
    if (child == NULL) {
        jac->setZero();
    }

    int idx = firstStateIdx;

    if (isStateAngle) {
        jac->set(0, idx, -(mpWorld.y - sy));
        jac->set(1, idx, (mpWorld.x - sx));
        idx++;
    }

    if (isStateLen) {
        double l;
        if (child == NULL) {
            // mpWorld is a point in this link's frame.  Therefore
            // dXdL, dYdL depend on the point's distance from the origin
            // along the x (length) axis
            ModelPoint mpLocal = worldToLocal(mpWorld);
            l = mpLocal.x;
        }
        else {
            // The effect on mpWorld of changing this link's length
            // depends on the position of the child link in this link's
            // frame
            l = (child->position.x);

        }
        jac->set(0, idx, l*ax);
        jac->set(1, idx, l*ay);
        idx++;
    }

    if (isStateThickness) {
        double t;
        if (child == NULL) {
            ModelPoint mpLocal = worldToLocal(mpWorld);
            t = mpLocal.y/2.0;
        }
        else {
            t = (child->position.y)/2.0;
        }
        jac->set(0, idx, -t*ay);
        jac->set(1, idx, t*ax);
        idx++;
    }

    if (isStatePos) {
        if (parent == NULL) {
            // This is the root link; position is in world frame
            jac->set(0, idx, 1);
            jac->set(1, idx, 0);
            jac->set(0, idx+1, 0);
            jac->set(1, idx+1, 1);
        }
        else {
            jac->set(0, idx, (parent->len)*(parent->ax));
            jac->set(1, idx, (parent->len)*(parent->ay));
            jac->set(0, idx+1, -(parent->thickness)/2.0*(parent->ay));
            jac->set(1, idx+1, (parent->thickness)/2.0*(parent->ax));
        }
    }

    if (parent != NULL) {
        parent->getKinematicJacobian(jac, mpWorld, this);
    }
}

/**
   Inserts the list of control points for this link, then recursively adds the
   control points for this link's descendants.
   @param controlPts list of control points to add to
 */
void SPMLink::getControlPts(vector<CvPoint> &controlPts) {
	if (canSetAngle || canSetLen) {
		controlPts.push_back(cvPoint(MathFun::round(ex), MathFun::round(ey)));
	}
	if (canSetThickness) {
		ModelPoint mp(.5, 1);
		mp = localToWorld(mp);
		controlPts.push_back(cvPoint(MathFun::round(mp.x), MathFun::round(mp.y)));
	}
	if (canSetPos) {
		controlPts.push_back(cvPoint(MathFun::round(sx), MathFun::round(sy)));
	}

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->getControlPts(controlPts);
	}
}

/**
    Inserts the starting point of this link into the list of joint centers.
    If this link has no children, also adds the ending point to the list.
    If there are children, recursively calls getJointCenters for the
    children.
    @param jointCtrs list of joint centers
 */
void SPMLink::getJointCenters(vector<CvPoint> &jointCtrs) {
    jointCtrs.push_back(cvPoint(MathFun::round(sx), MathFun::round(sy)));

    if (children.size() == 0) {
        jointCtrs.push_back(cvPoint(MathFun::round(ex), MathFun::round(ey)));
    }
    else {
        vector<SPMLink*>::iterator itr = children.begin();
        for (; itr != children.end(); itr++) {
            (*itr)->getJointCenters(jointCtrs);
        }
    }
}

/**
   Sets the state of this link from the specified control points, then 
   recursively sets the states of its descendants.
   @param controlPt control point that was moved
   @param idx index of first state element of this link's state in the
        state vector
    @param cidx index of control point that was moved
    @return true if this link is the one whose state has changed

 */
bool SPMLink::setStateFromControlPts(CvPoint controlPt, int &idx,
									 int cidx) {
	bool done = false;

	if (canSetAngle || canSetLen) {
		if (idx == cidx) {
			if (canSetAngle) {
				angle = atan2(controlPt.y - sy, controlPt.x - sx);
				if (parent != NULL) {
					localAngle = angle - parent->angle;
				}
				else {
					localAngle = angle;
				}
			}
			if (canSetLen) {
				len = MathFun::dist(controlPt, cvPoint(sx, sy));
			}
			done = true;
		}
		else {
			idx++;
		}
	}
	if (!done && canSetThickness) {
		if (idx == cidx) {
			ModelPoint mp = localToWorld(ModelPoint(.5, 0));
			thickness = (2.0)*MathFun::dist(controlPt, cvPoint(mp.x, mp.y));
			done = true;
		}
		else {
			idx++;
		}
	}
	if (!done && canSetPos) {
		if (idx == cidx) {
			if (parent == NULL) {
				position.x = controlPt.x;
				position.y = controlPt.y;
			}
			else {
                position= parent->worldToLocal(
                    ModelPoint(controlPt.x, controlPt.y));
			}
			done = true;
		}
		else {
			idx++;
		}
	}


	if (done) {
		// Since this link's state has changed, we need to propagate the
		// change to its descendants
		calcEndPts(true);
		return true;
	}
	else {
		vector<SPMLink*>::iterator itr = children.begin();
		for (; itr != children.end(); itr++) {
			done = (*itr)->setStateFromControlPts(controlPt, idx, cidx);
			if (done) {
				break;
			}
		}
		return done;
	}
}

/*
    @param mpLocal coordinates of a point, in the link's local coordinate
        system
    @return true if the point, given in local coordinates, is inside
        this link.
 */
bool SPMLink::contains(ModelPoint &mpLocal) const {
    return (mpLocal.x >= 0 && mpLocal.x <= 1 
            && mpLocal.y <= 1 && mpLocal.y >= -1);
}

/**
   Draws a solid, colored rectangle representing this link, then recursively
   draws its descendant links.
   @param image image to draw on
   @param color color in which to draw
 */
void SPMLink::drawFilled(IplImage *image, CvScalar color) {
    CvPoint *corners = calcCorners();
    cvFillConvexPoly(image, corners, 4, color);
    delete [] corners;

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->drawFilled(image, color);
	}
}

/**
   Draws the outline of this link's bounding box, then recursively
   draws its descendant links.
   @param image image to draw on
   @param color color in which to draw
 */
void SPMLink::drawBoxes(IplImage *image, CvScalar color) {
    CvPoint *corners = calcCorners();
    int npts = 4;
    cvPolyLine(image, &corners, &npts, 1, 1, color);
    delete [] corners;

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->drawBoxes(image, color);
	}
}

/**
   Draws this link's axis, then recursively
   draws its descendant links.
   @param image image to draw on
   @param color color in which to draw
 */
void SPMLink::drawLines(IplImage *image, CvScalar color) {
    int radius = 8;

    cvLine(image, cvPoint(sx, sy), cvPoint(ex, ey), color, 4);
    cvCircle(image, cvPoint(sx, sy), radius, color, -1);
    cvCircle(image, cvPoint(ex, ey), radius, color, -1);

	vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->drawLines(image, color);
	}
}

/*
    Draws the link as a stick and two balls.  For this function,
    ignore the stick's thickness and use instead r, the radius
    of the balls.
    @param image image to draw on
    @param color color in which to draw
    @param r radius of ball
 */
void SPMLink::drawSticksAndBalls(IplImage *image, CvScalar color, int r) {
    cvCircle(image, cvPoint(sx, sy), r, color, -1);
    cvCircle(image, cvPoint(ex, ey), r, color, -1);
    
    cvLine(image, cvPoint(sx, sy), cvPoint(ex, ey), color, r/4, CV_AA);

    vector<SPMLink*>::iterator itr = children.begin();
	for (; itr != children.end(); itr++) {
		(*itr)->drawSticksAndBalls(image, color, r);
	}
}


/**
    Given an edge image, extract the rectangular portion that corresponds
        to this link.  This function is used in the edge appearance model
        initialization process.  It extracts a small rectangle region of
        the image that includes the area of the image corresponding to
        this link.  The link is outlined in gray.
    @param edgeImg single channel edge image
    @param origin  the value returned is the origin of the linkImg in
        edgeImg
    @param linkImg the image returned is the rectangular part of edgeImg
        which contains this link.  linkImg is a single channel image
        with three possible values:
        - 0 -- no edge
        - 128 -- link boundary
        - 255 -- edge
 */
void SPMLink::getLinkEdgeImg(IplImage *edgeImg, CvPoint &origin, 
                             IplImage *&linkImg) const {
    int minx, maxx, miny, maxy;
    CvRect roi;
    IplImage *tmpImg;

    // Make a copy of input image that I can draw on w/o changing the original
    tmpImg = ImgFun::newImage(cvSize(edgeImg->width, edgeImg->height),
                              IPL_DEPTH_8U, 1);
    cvCopy(edgeImg, tmpImg);

    // Draw the link on the temporary image
    CvPoint *corners = calcCorners();
    int npts = 4;
    cvPolyLine(tmpImg, &corners, &npts, 1, 1, cvScalar(128));
    delete [] corners;

    // Calculate the rectangular region that contains this link
    corners = calcCorners();
    minx = miny = tmpImg->height + 1;
    maxx = maxy = -1;
    for (int k = 0; k < 4; k++) {
        if (corners[k].x < minx) {
            minx = corners[k].x;
        }
        if (corners[k].x > maxx) {
            maxx = corners[k].x;
        }
        
        if (corners[k].y < miny) {
            miny = corners[k].y;
        }
        if (corners[k].y > maxy) {
            maxy = corners[k].y;
        }
    }
    roi.x = minx;
    roi.y = miny;
    roi.width = maxx - minx + 1;
    roi.height = maxy - miny + 1;

    // Extract the interesting region from tmpImg
    cvSetImageROI(tmpImg, roi);
    linkImg = ImgFun::newImage(cvSize(roi.width, roi.height), IPL_DEPTH_8U, 1);
    cvCopy(tmpImg, linkImg);
    origin.x = roi.x;
    origin.y = roi.y;
    
    cvReleaseImage(&tmpImg);
	delete [] corners;
}

/**
    Given a grayscale image, extract the rectangular portion that corresponds
        to this link.  This function is used in the patch appearance model
        initialization process.  It extracts a small rectangle region of
        the image that includes the area of the image corresponding to
        this link.  The link is outlined in gray.
    @param patchImg grayscale image
    @param origin the value returned is the origin of the linkImg in
        patchImg
    @param linkImg the image returned is the rectangular part of patchImg
        which contains this link.  linkImg is a single channel image
 */
void SPMLink::getLinkPatchImg(IplImage *patchImg, CvPoint &origin, 
                             IplImage *&linkImg) const {
    int minx, maxx, miny, maxy;
    CvRect roi;
    IplImage *tmpImg;

    // Make a copy of input image that I can draw on w/o changing the original
    tmpImg = ImgFun::newImage(cvSize(patchImg->width, patchImg->height),
                              IPL_DEPTH_8U, 1);
    cvCopy(patchImg, tmpImg);

    // Draw the link on the temporary image
    CvPoint *corners = calcCorners();
    int npts = 4;
    cvPolyLine(tmpImg, &corners, &npts, 1, 1, cvScalar(128));
    delete [] corners;

    // Calculate the rectangular region that contains this link
    corners = calcCorners();
    minx = miny = tmpImg->height + 1;
    maxx = maxy = -1;
    for (int k = 0; k < 4; k++) {
        if (corners[k].x < minx) {
            minx = corners[k].x;
        }
        if (corners[k].x > maxx) {
            maxx = corners[k].x;
        }
        
        if (corners[k].y < miny) {
            miny = corners[k].y;
        }
        if (corners[k].y > maxy) {
            maxy = corners[k].y;
        }
    }
    roi.x = minx;
    roi.y = miny;
    roi.width = maxx - minx + 1;
    roi.height = maxy - miny + 1;

    // Extract the interesting region from tmpImg
    cvSetImageROI(tmpImg, roi);
    linkImg = ImgFun::newImage(cvSize(roi.width, roi.height), IPL_DEPTH_8U, 1);
    cvCopy(tmpImg, linkImg);
    origin.x = roi.x;
    origin.y = roi.y;
    
    cvReleaseImage(&tmpImg);
	delete [] corners;
}

/**
   Calculate the corners of this link's bounding box, in world coordinates.
   Return them in the order bottom left, bottom right, top right, 
   top left.  "Bottom" is defined to be the end at the link origin.
   @return pointer to an array of 4 CvPoints -- the locations of 
        this link's corners
 */
CvPoint *SPMLink::calcCorners() const {
    CvPoint *corners = new CvPoint[4];
    double x, y;

    // bottom left
    x = sx - ay*thickness/2.0;
    y = sy + ax*thickness/2.0;
    corners[0] = cvPoint(MathFun::round(x), MathFun::round(y));

    // bottom right
    x = sx + ay*thickness/2.0;
    y = sy - ax*thickness/2.0;
    corners[1] = cvPoint(MathFun::round(x), MathFun::round(y));

    // top right
    x = ex + ay*thickness/2.0;
    y = ey - ax*thickness/2.0;
    corners[2] = cvPoint(MathFun::round(x), MathFun::round(y));

    // top left
    x = ex - ay*thickness/2.0;
    y = ey + ax*thickness/2.0;
    corners[3] = cvPoint(MathFun::round(x), MathFun::round(y));

    return corners;
}


/**
   Transforms the given point from local to world coordinates.
   @param mpLocal location of the point in local coordinates
   @return location of the point in world coordinates
 */
ModelPoint SPMLink::localToWorld(const ModelPoint &mpLocal) const {
    ModelPoint mpWorld;
    mpWorld.x = len*(mpLocal.x)*cos(angle) 
                - (thickness/2.0)*(mpLocal.y)*sin(angle) + sx;
    mpWorld.y = len*(mpLocal.x)*sin(angle) 
                + (thickness/2.0)*(mpLocal.y)*cos(angle) + sy;
    return mpWorld;    
}

/**
   Transforms the given point from world to local coordinates.
   @param mpWorld location of the point in world coordinates
   @return location of the point in local coordinates
 */
ModelPoint SPMLink::worldToLocal(const ModelPoint &mpWorld) const {
    ModelPoint mpLocal;
	mpLocal.x = ((mpWorld.x - sx)*cos(-angle) - (mpWorld.y - sy)*sin(-angle))
				/len;
    mpLocal.y = ((mpWorld.x - sx)*sin(-angle) + (mpWorld.y - sy)*cos(-angle))
				/(thickness/2.0);
    return mpLocal;
}