#ifndef SPMLINK_INCLUDED
#define SPMLINK_INCLUDED

#include <vector>
#include <iostream>
#include <fstream>
using namespace std;

#include "cv.h"
#include "Vector.h"
#include "Matrix.h"
#include "ModelPoint.h"

#define MAX_LINE_LEN (160)

/** @addtogroup kin_group */
/* @{ */
/** A scaled prismatic model (SPM) link in the kinematic model */
class SPMLink {
private:
    /** Name of this link */
    char *name;

    /** World coordinates of origin of this link */
    double sx, sy; 

    /** World coordinates of endpoint of this link */
    double ex, ey; 

    /** Axis; unit vector pointing from (sx,sy) to (ex,ey) */
    double ax, ay; 

    /** Rotation of this link, relative to the parent's frame */
    double localAngle;

    /** Rotation of this link, relative to the world frame */
	double angle;

    /** Length of this link */
    double len;

    /** Thickness of the link */
    double thickness;

    /** Position of this link's origin in its parent's frame */
	ModelPoint position;

    /** These flags are true if the corresponding variables 
	     are part of the state vector */
	bool isStateAngle, isStateLen, isStateThickness, isStatePos;

	/** These flags are true if the corresponding variables
	     can be set by the user via control points */
	bool canSetAngle, canSetLen, canSetThickness, canSetPos;

    /** First and last indices where this link's state appears in the
        kinematic state vector */
    int firstStateIdx, lastStateIdx;

    /** Pointer to this link's parent in the kinematic tree */
	SPMLink *parent;

    /** List of pointers to this link's children in the kinematic tree */
	vector<SPMLink*> children;

	void calcEndPts(bool recurse);

	void readFromFile(ifstream &fin, int &firstIdx, char *pname = NULL);

public:
    SPMLink(ifstream &fin);
    SPMLink(SPMLink *parent, ifstream &fin, int &firstIdx, char *pname = NULL);
	~SPMLink();

    int countStates();
    void listLinks(vector<SPMLink *> &v);
	bool setState(const Vector &v, int &idx);
	void getState(Vector &v, int &idx);
    void getKinematicJacobian(Matrix *jac, ModelPoint mpWorld, 
                              const SPMLink *child) const;
	void getControlPts(vector<CvPoint> &controlPts);
    void getJointCenters(vector<CvPoint> &jointCtrs);
	bool setStateFromControlPts(CvPoint controlPt, int &idx, 
		                        int cidx);

    CvPoint *calcCorners() const;

    bool contains(ModelPoint &mpLocal) const;

	void drawFilled(IplImage *image, CvScalar color);
	void drawBoxes(IplImage *image, CvScalar color);
	void drawLines(IplImage *image, CvScalar color);
    void drawSticksAndBalls(IplImage *image, CvScalar color, int r);
	
    void getLinkEdgeImg(IplImage *edgeImg, CvPoint &origin,
                        IplImage *&linkImg) const;
    void getLinkPatchImg(IplImage *patchImg, CvPoint &origin,
                        IplImage *&linkImg) const;

    void writeToFile(ofstream &fout);

    ModelPoint localToWorld(const ModelPoint &mpLocal) const;
	ModelPoint worldToLocal(const ModelPoint &mpWorld) const;
};
/* @} */


#endif // !def SPMLINK_INCLUDED
