#include "cv.h"
#include "Vector.h"
#include <iostream>
using namespace std;

#define NULL 0

/*
    Constructors & destructor
 */
Vector::Vector() : vector(NULL) { }

Vector::Vector(int len) : vector(NULL) {
    initialize(len);
}

Vector::Vector(const Vector &v) : vector(NULL) {
    duplicate(v);
}

Vector::~Vector() {
    destroy();
}

/**
    Sets the vector length and allocates memory.
 */
void Vector::initialize(int len) {
    assert(len > 0);
    destroy();
    vector = cvCreateMat(len, 1, CV_64FC1);
}

void Vector::destroy() {
    if (vector != NULL) {
        cvReleaseMat(&vector);
    }
    vector = NULL;
}

/*
    Accessors
 */
/**
    @return length of vector
 */
int Vector::len() const {
    if (vector == NULL) {
        return 0;
    }
    else {
        return vector->rows;
    }
}

/**
    Gets an element of the vector
    @param i zero-based index
    @return element i of the vector
 */
double Vector::get(int i) const {
    assert(vector != NULL);

    return cvmGet(vector, i, 0);
}

/**
    Sets an element of the vector
    @param i zero-based index
    @param d new value of element i
 */
void Vector::set(int i, double d) {
    assert (vector != NULL);

    cvmSet(vector, i, 0, d);
}

/**
    Sets all elements of the vector to zero
 */
void Vector::setZero() {
    assert(vector != NULL);

    cvSetZero(vector);
}

/**
    Sets every element to its opposite.
 */
void Vector::negate() {
    assert(vector != NULL);
    
    for (int i = 0; i < len(); i++) {
        set(i, -get(i));
    }
}

/**
    @return true if every element in the vector is zero
 */
bool Vector::isZero() {
    assert (vector != NULL);

    for (int i = 0; i < len(); i++) {
        if (get(i) != 0) {
            return false;
        }
    }

    return true;
}

/*
   Operations
 */
Vector Vector::operator=(const Vector &v) {
    duplicate(v);
    
    return *this;
}

Vector Vector::operator-() {
    Vector retval(len());
    
    for (int i = 0; i < len(); i++) {
        retval.set(i, -get(i));
    }

    return retval;
}

Vector Vector::operator+(const Vector &v) const {
    assert (len() == v.len());

    Vector retval(len());
    cvAdd(vector, v.vector, retval.vector);
    return retval;
}

void Vector::operator+=(const Vector &v) {
    assert(len() == v.len());

    cvAdd(vector, v.vector, vector);
}

Vector Vector::operator-(const Vector &v) const {
    assert (len() == v.len());

    Vector retval(len());
    cvSub(vector, v.vector, retval.vector);
    return retval;
}

/** @addtogroup util_group */
/* @{ */
/**
    Checks for equality between two vectors.
    @param v1 vector 1
    @param v2 vector 2
    @return true if the two vectors are the same
 */
bool operator==(const Vector &v1, const Vector &v2) {
    assert(v1.len() == v2.len());

    for (int i = 0; i < v1.len(); i++) {
        if (v1(i) != v2(i)) {
            return false;
        }
    }

    return true;
}

/**
    Checks for inequality between two vectors.
    @param v1 vector 1
    @param v2 vector 2
    @return false if the two vectors are the same
 */
bool operator!=(const Vector &v1, const Vector &v2) {
    return !(v1 == v2);
}
/* @} */


/**
   Calculates dot product of two vectors.
   @param v vector
   @return dot product of this vector with v
 */
double Vector::operator*(const Vector &v) const {
    assert (len() == v.len());
    
    return cvDotProduct(vector, v.vector);
}


/*
   Multiplication by a scalar
 */
Vector Vector::operator*(double s) const {
    Vector retval(len());

    for (int i = 0; i < len(); i++) {
        retval.set(i, this->get(i) * s);
    }

    return retval;
}

Vector Vector::vectorproduct(const Vector &v)
{
	assert(v.len() == len());

	Vector retval(len());
	for (int i = 0; i < len(); i++) {
		retval.set(i, this->get(i) * v.get(i));
	}
	
	return retval;
}

void Vector::duplicate(const Vector &v) {    
    destroy();

    if (v.vector == NULL) {
        vector = NULL; 
    }
    else {
        vector = cvCloneMat(v.vector);
    }
}