#include "cv.h"
//#include "ipl.h"
#include "highgui.h"

#include <cmath>

#include "EdgePixelGrabber.h"
#include "ImgFun.h"
#include "MathFun.h"

EdgePixelGrabber *EdgePixelGrabber::theEPG = NULL;    // Initialize pointer

/** 
    Constructor 
 */
EdgePixelGrabber::EdgePixelGrabber() { 
    BOX_COLOR = CV_RGB(0, 255, 0);
    CP_COLOR = CV_RGB(255, 0, 0);
    
    kt = NULL;
    eam = NULL;
    origImg = NULL;
    displayImg = NULL;
}

/**
    Destructor
 */
EdgePixelGrabber::~EdgePixelGrabber() {
    cvReleaseImage(&origImg);
    cvReleaseImage(&edgeImg);
    cvReleaseImage(&displayImg);
}

/** 
    Returns a pointer to the instance of this class.
    Don't call the constructor to get an instance of
    this class; instead call this static method.
    @return pointer to the single instance of this class
 */
EdgePixelGrabber * EdgePixelGrabber::getInstance() {
    if (theEPG == NULL) {
        // First call; create instance
        theEPG = new EdgePixelGrabber();
    }
    return theEPG;
}

/**
    Sets the kinematic model, appearance model, and image to use 
    @param kt kinematic model
    @param eam appearance model
    @param img image to use to initialize edges
 */
void EdgePixelGrabber::initialize(KinematicTree *kt, EdgeAppearanceModel *eam,
                                  IplImage *img) {
    this->kt = kt;
    this->eam = eam;
    mode = 0;
    origImg = ImgFun::newImage(cvSize(img->width, img->height), img->depth,
                               img->nChannels);
    cvCopy(img, origImg);
    
    edgeImg = eam->getEdgeImage(origImg);
    cvGetRawData(edgeImg, (uchar**)&edgeData, &edgeStep, &edgeSize);
    edgeStep /= sizeof(edgeData[0]);

    display();
}

/**
   Resets the edge image to what it originally was
 */
void EdgePixelGrabber::reset() {
    edgeImg = eam->getEdgeImage(origImg);
    cvGetRawData(edgeImg, (uchar**)&edgeData, &edgeStep, &edgeSize);
    edgeStep /= sizeof(edgeData[0]);
}

/**
   Advance to the next stage in the process.  
   @return true for success, false if currently on the final step.
 */
bool EdgePixelGrabber::nextStep() {
    if (mode == kt->countLinks()) {
        return false;
    }
    else {
        mode++;
        display();
        return true;
    }
}

/**
   Moves back to the previous stage in the process.  
   @return true for success, false if currently on the first step.
 */
bool EdgePixelGrabber::prevStep() {
    if (mode == 0) {
        return false;
    }
    else {
        mode--;
        display();
        return true;
    }
}

/**
    Retrieves the link images and their origins for all links.
    This is called after all stages of initialization are complete,
    to get the information needed for initializing the appearance
    model.
    @param linkImgs array of images, one per link, where the edge
        pixel images for each link are returned
    @param origins array of points where the origin of each link is
        returned
 */
void EdgePixelGrabber::getData(IplImage **linkImgs, CvPoint *origins) {
    int n = kt->countLinks();
    
    for (int i = 0; i < n ; i++) {
        (*kt)[i]->getLinkEdgeImg(edgeImg, origins[i], linkImgs[i]);
    }
}

/**
   Opens the display window and shows the image corresponding to the 
   current stage.
 */
void EdgePixelGrabber::display() {
    if (mode == 0) {
        // display the full-body image
        if (displayImg != NULL) {
            cvReleaseImage(&displayImg);
        }
        displayImg = ImgFun::newImage(cvSize(edgeImg->width, edgeImg->height));
        cvConvertImage(edgeImg, displayImg);

        // Boxes around links
        kt->drawBoxes(displayImg, BOX_COLOR);

        // Control points
        controlPts.erase(controlPts.begin(), controlPts.end());
        kt->getControlPts(controlPts);
        vector<CvPoint>::iterator itr = controlPts.begin();
	    for (; itr != controlPts.end(); itr++) {
		    cvCircle(displayImg, (*itr), CP_RADIUS, CP_COLOR, CV_FILLED);
	    }

        // Set up the window and appropriate mouse handler
        cvNamedWindow(EPG_WINDOW_NAME, CV_WINDOW_AUTOSIZE);
        dragging = false;
        cvSetMouseCallback(EPG_WINDOW_NAME, EPGMode0MouseCallback, 0);
        cvShowImage(EPG_WINDOW_NAME, displayImg);
    }
    else if (mode >= 1 && mode <= kt->countLinks()){
        // Get the enlarged link image for the link and display it
        IplImage *linkImg;
        (*kt)[mode-1]->getLinkEdgeImg(edgeImg, origin, linkImg);
        if (displayImg != NULL) {
            cvReleaseImage(&displayImg);
        }
        displayImg = enlargeImage(linkImg, SCALE);
        cvReleaseImage(&linkImg);

        // Set up the window and appropriate mouse handler
        cvNamedWindow(EPG_WINDOW_NAME, CV_WINDOW_AUTOSIZE);
        cvSetMouseCallback(EPG_WINDOW_NAME, EPGModeNMouseCallback, 0);
        cvShowImage(EPG_WINDOW_NAME, displayImg);
    }
}


/**
    Closes the display window.
 */
void EdgePixelGrabber::hide() {
    cvDestroyWindow(EPG_WINDOW_NAME);
}

/**
   Takes in a grayscale image, returns the image scaled up scale times.
   This function treats the input image as a binary image.
   @param src input image
   @param scale factor to multiply size of each pixel by
   @return pointer to the enlarged image
 */
IplImage * EdgePixelGrabber::enlargeImage(IplImage *src, int scale) {
    if (scale < 2) {
        return NULL;
    }
    if (src->depth != IPL_DEPTH_8U || src->nChannels != 1) {
        return NULL;
    }

    IplImage *bigger = cvCreateImage(cvSize(src->width*scale, src->height*scale),
                                     IPL_DEPTH_8U, 1);
    bigger->origin = src->origin;
    
    unsigned char *data;
    int step;
    CvSize size;

    cvGetRawData(src, (uchar**)&data, &step, &size);
    step /= sizeof(data[0]);

    cvRectangle(bigger, cvPoint(0, 0), 
                cvPoint(bigger->width-1, bigger->height-1), 
                cvScalar(0), CV_FILLED);

    for (int x = 0; x < src->width; x++) {
        for (int y = 0; y < src->height; y++) {
            if (data[y*step + x]) {
                cvRectangle(bigger, cvPoint(scale*x, scale*y), 
                            cvPoint(scale*(x+1) - 1, scale*(y+1) - 1), 
                            cvScalar(data[y*step + x]), CV_FILLED);
            }
        }
    }

    return bigger;
}

/**
    Mouse handler used in stage 1.  Handles positioning the kinematic
    model over the input image.  This should not be called directly.
    It is called by the mouse handler for the display window.
    @param event the type of mouse event
    @param x x coordinate of mouse event
    @param y y coordinate of mouse event
    @param flags mouse event flags
 */
void EdgePixelGrabber::mode0Mouse(int event, int x, int y, int flags) {
	CvPoint clickPoint = cvPoint(x, displayImg->height - y);
    int i;
    vector<CvPoint>::iterator itr;

	switch(event) {
    case CV_EVENT_LBUTTONDOWN:
		itr = controlPts.begin();
		for (i = 0; itr != controlPts.end(); itr++, i++) {
            if (MathFun::dist((*itr), clickPoint) <= CP_RADIUS) {
                draggedIdx = i;
                dragging = true;
                break;
            }
        }
        break;
    
    case CV_EVENT_LBUTTONUP:
    case CV_EVENT_MOUSEMOVE:
        if (dragging) {
            kt->setStateFromControlPts(clickPoint, draggedIdx);
            controlPts.erase(controlPts.begin(), controlPts.end());
			kt->getControlPts(controlPts);
			
			// Re-draw image
            cvConvertImage(edgeImg, displayImg);
            kt->drawBoxes(displayImg, BOX_COLOR);

			vector<CvPoint>::iterator itr = controlPts.begin();
			for (; itr != controlPts.end(); itr++) {
				cvCircle(displayImg, (*itr), CP_RADIUS, CP_COLOR, CV_FILLED);
			}
			cvShowImage(EPG_WINDOW_NAME, displayImg);
        }
        if (event == CV_EVENT_LBUTTONUP) {
            dragging = false;
        }
        break;

    default:
        break;
    }
}

/**
    Mouse handler used in stages after stage 1.  Handles turning individual
    pixels on and off. Toggles the clicked pixel between
    255 (edge) and 0 (non-edge).
    @param event the type of mouse event
    @param x x coordinate of mouse event
    @param y y coordinate of mouse event
    @param flags mouse event flags
 */
void EdgePixelGrabber::modeNMouse(int event, int x, int y, int flags) {
    if (event != CV_EVENT_LBUTTONUP) {
        return;
    }

    // Translate coordinates from linkImg to edgeImg
    int edgex = (int)floor((double)x / SCALE) + origin.x;
    int edgey = displayImg->height/SCALE - 1 - (int)floor((double)y / SCALE) 
                + origin.y;

    if (edgeData[edgey*edgeStep + edgex] == 255) {
        edgeData[edgey*edgeStep + edgex] = 0;
        cvRectangle(displayImg, 
            cvPoint(SCALE*(edgex-origin.x), SCALE*(edgey-origin.y)), 
            cvPoint(SCALE*(edgex-origin.x+1)-1, SCALE*(edgey-origin.y+1)-1),
            cvScalar(0), CV_FILLED);
    }
    else if (edgeData[edgey*edgeStep + edgex] == 0) {
        edgeData[edgey*edgeStep + edgex] = 255;
        cvRectangle(displayImg, 
            cvPoint(SCALE*(edgex-origin.x), SCALE*(edgey-origin.y)), 
            cvPoint(SCALE*(edgex-origin.x+1)-1, SCALE*(edgey-origin.y+1)-1),
            cvScalar(255), CV_FILLED);
    }

    cvShowImage(EPG_WINDOW_NAME, displayImg);
}


/** @addtogroup bridge_group */
/* @{ */
/**
    Mouse callback function used by the HighGUI window when in 
    stage 1.  This just calls EdgePixelGrabber::mode0Mouse().
    @param event the type of mouse event
    @param x x coordinate of mouse event
    @param y y coordinate of mouse event
    @param flags mouse event flags
*/
void EPGMode0MouseCallback(int event, int x, int y, int flags, void* param) {
    EdgePixelGrabber::getInstance()->mode0Mouse(event, x, y, flags);
}

/**
    Mouse callback function used by the HighGUI window when in 
    stages after stage 1.  This just calls 
    EdgePixelGrabber::mode0Mouse().
    @param event the type of mouse event
    @param x x coordinate of mouse event
    @param y y coordinate of mouse event
    @param flags mouse event flags
*/
void EPGModeNMouseCallback(int event, int x, int y, int flags, void* param) {
    EdgePixelGrabber::getInstance()->modeNMouse(event, x, y, flags);
}
/* @} */