#ifndef EDGE_PIXEL_GRABBER_INCLUDED
#define EDGE_PIXEL_GRABBER_INCLUDED

#include <vector>
using namespace std;

#include "cv.h"
#include "highgui.h"

#include "KinematicTree.h"
#include "EdgeAppearanceModel.h"

#define EPG_WINDOW_NAME ("epg_window")

/** @defgroup gui_group GUI */

/** @defgroup bridge_group GUI Bridge 
    @ingroup gui_group */
/* @{ */
/**
    Handles collecting edge pixels to initialize an edge 
    appearance model. 
    
    There are several stages in the initialization.  This class
    handles a HighGUI window and mouse input for each stage.
    Use the nextStep() and prevStep() to go forward and backward
    through the stages.
    In the first stage, the window displays edges from a given
    image.  The user can use the mous to position the links of 
    the kinematic model over the corresponding regions of the image.
    After this is complete, there is one stage per link.  In each
    of these stages, an enlarged view of the pixels for the 
    corresponding link is shown.  The user can use the mouse to
    turn individual edge pixels on and off.  After all stages are
    complete, it has the edges associated with each link.

    This is a singleton class.  There is only one instance of
    this class.  To get a pointer to the class, call the
    getInstance() method.  Don't call the constructor.
    The singleton design pattern is used because the mouse
    callback functions have to be able to call into this
    function.
 */
class EdgePixelGrabber {
private:
    /** Pointer to the one and only instance of this class */
    static EdgePixelGrabber *theEPG;

    /** The kinematic model */
    KinematicTree *kt;

    /** The appearance model */
    EdgeAppearanceModel *eam;

    /** The original background image being used for model initialization */
    IplImage *origImg;
    
    /** Edge transform of origImg */
    IplImage *edgeImg;
    
    /** The image currently being displayed in the output window */
    IplImage *displayImg;
    
    /** Raw pixel data from the edge image */
    unsigned char *edgeData; 

    /** Length of a row of raw edge image pixel data in bytes */
    int edgeStep; 

    /** Size of edge image */
    CvSize edgeSize;

    /** Control points; used to manipulate kinematic model state with mouse */
    vector<CvPoint> controlPts;

    /** The origin of the link for this stage */
    CvPoint origin;

    /** Mode of operation; 0 in first stage, 1 in later stages */
    int mode;

    /** True if and only if a control point is currently being dragged */
    bool dragging;

    /** Index of control point currently being dragged */
    int draggedIdx;
    
    /** Color of boxes around links */
    CvScalar BOX_COLOR;

    /** Color used to draw control points */
    CvScalar CP_COLOR;
    
    /** Radius of control points */
    enum {CP_RADIUS = 5};       

    /** Scale factor of enlarged images used in stages after stage 1 */
    enum {SCALE = 8};

    EdgePixelGrabber();
    EdgePixelGrabber(const EdgePixelGrabber &epg);
    EdgePixelGrabber & operator=(const EdgePixelGrabber &epg);
    ~EdgePixelGrabber();
    void display();
    IplImage * enlargeImage(IplImage *src, int scale);

public:
    static EdgePixelGrabber * getInstance();
    void initialize(KinematicTree *kt, EdgeAppearanceModel *eam,
                    IplImage *img);
    void reset();

    bool nextStep();
    bool prevStep();
    void getData(IplImage **linkImgs, CvPoint *origins);
    void hide();

    void mode0Mouse(int event, int x, int y, int flags);
    void modeNMouse(int event, int x, int y, int flags);
}; 
/* @} */

// Mouse callback functions
void EPGMode0MouseCallback(int event, int x, int y, int flags, void* param);
void EPGModeNMouseCallback(int event, int x, int y, int flags, void* param);

#endif // !def EDGE_PIXEL_GRABBER_INCLUDED
