#include "cv.h"
//#include "ipl.h"
#include "highgui.h"

#include <cmath>

#include "PatchPixelGrabber.h"
#include "ImgFun.h"
#include "MathFun.h"

PatchPixelGrabber *PatchPixelGrabber::thePPG = NULL;    // Initialize pointer

PatchPixelGrabber::PatchPixelGrabber() { 
    BOX_COLOR = CV_RGB(0, 255, 0);
    CP_COLOR = CV_RGB(255, 0, 0);
    BOUNDARY_COLOR = cvScalar(128);
    
    kt = NULL;
    pam = NULL;
    origImg = NULL;
    displayImg = NULL;
}

PatchPixelGrabber::~PatchPixelGrabber() {
    cvReleaseImage(&origImg);
    cvReleaseImage(&patchImg);
    cvReleaseImage(&displayImg);
}

PatchPixelGrabber * PatchPixelGrabber::getInstance() {
    if (thePPG == NULL) {
        // First call; create instance
        thePPG = new PatchPixelGrabber();
    }
    return thePPG;
}

void PatchPixelGrabber::initialize(KinematicTree *kt, PatchAppearanceModel *pam,
                                  IplImage *img) {
    this->kt = kt;
    this->pam = pam;
    mode = 0;
    origImg = ImgFun::newImage(cvSize(img->width, img->height), img->depth,
                               img->nChannels);
    cvCopy(img, origImg);

    patchImg = ImgFun::newImage(cvSize(img->width, img->height), img->depth, 1);
    cvConvertScale(origImg, patchImg);
    
    cvGetRawData(patchImg, (uchar**)&patchData, &patchStep, &patchSize);
    patchStep /= sizeof(patchData[0]);

    display();
}

/*
   Reset the patch image to what it originally was
 */
void PatchPixelGrabber::reset() {
    patchImg = ImgFun::newImage(cvSize(origImg->width, origImg->height), 
                                origImg->depth, 1);
    cvConvertScale(origImg, patchImg);
    cvGetRawData(patchImg, (uchar**)&patchData, &patchStep, &patchSize);
    patchStep /= sizeof(patchData[0]);
}

/*
    Retrieve the link images and their origins from all links
 */
void PatchPixelGrabber::getData(IplImage **linkImgs, CvPoint *origins) {
    int n = kt->countLinks();
    
    for (int i = 0; i < n ; i++) {
        (*kt)[i]->getLinkPatchImg(patchImg, origins[i], linkImgs[i]);
    }
}

/*
   Show the image corresponding to the current mode.
 */
void PatchPixelGrabber::display() {
    if (mode == 0) {
        // display the full-body image
        if (displayImg != NULL) {
            cvReleaseImage(&displayImg);
        }
        displayImg = ImgFun::newImage(cvSize(patchImg->width, patchImg->height));
        cvConvertImage(patchImg, displayImg);

        // Boxes around links
        kt->drawBoxes(displayImg, BOX_COLOR);

        // Control points
        controlPts.erase(controlPts.begin(), controlPts.end());
        kt->getControlPts(controlPts);
        vector<CvPoint>::iterator itr = controlPts.begin();
	    for (; itr != controlPts.end(); itr++) {
		    cvCircle(displayImg, (*itr), CP_RADIUS, CP_COLOR, CV_FILLED);
	    }

        // Set up the window and appropriate mouse handler
        cvNamedWindow(PPG_WINDOW_NAME, CV_WINDOW_AUTOSIZE);
        dragging = false;
        cvSetMouseCallback(PPG_WINDOW_NAME, PPGMode0MouseCallback, 0);
        cvShowImage(PPG_WINDOW_NAME, displayImg);
    }
}


void PatchPixelGrabber::hide() {
    cvDestroyWindow(PPG_WINDOW_NAME);
}


void PatchPixelGrabber::mode0Mouse(int event, int x, int y, int flags) {
	CvPoint clickPoint = cvPoint(x, displayImg->height - y);
    int i;
    vector<CvPoint>::iterator itr;

	switch(event) {
    case CV_EVENT_LBUTTONDOWN:
		itr = controlPts.begin();
		for (i = 0; itr != controlPts.end(); itr++, i++) {
            if (MathFun::dist((*itr), clickPoint) <= CP_RADIUS) {
                draggedIdx = i;
                dragging = true;
                break;
            }
        }
        break;
    
    case CV_EVENT_LBUTTONUP:
    case CV_EVENT_MOUSEMOVE:
        if (dragging) {
            kt->setStateFromControlPts(clickPoint, draggedIdx);
            controlPts.erase(controlPts.begin(), controlPts.end());
			kt->getControlPts(controlPts);
			
			// Re-draw image
            cvConvertImage(patchImg, displayImg);
            kt->drawBoxes(displayImg, BOX_COLOR);

			vector<CvPoint>::iterator itr = controlPts.begin();
			for (; itr != controlPts.end(); itr++) {
				cvCircle(displayImg, (*itr), CP_RADIUS, CP_COLOR, CV_FILLED);
			}
			cvShowImage(PPG_WINDOW_NAME, displayImg);
        }
        if (event == CV_EVENT_LBUTTONUP) {
            dragging = false;
        }
        break;

    default:
        break;
    }
}


/** @addtogroup bridge_group */
/* @{ */
// External mouse callback functions
void PPGMode0MouseCallback(int event, int x, int y, int flags, void* param) {
    PatchPixelGrabber::getInstance()->mode0Mouse(event, x, y, flags);
}
/* @} */
