#include "cv.h"
#include "highgui.h"

#include "AppearanceModel.h"
#include "ImgFun.h"
#include "KinematicTree.h"
#include "StateSetter.h"
#include "MathFun.h"

StateSetter *StateSetter::theStateSetter = NULL;    // Initialize pointer

/** Constructor */
StateSetter::StateSetter() {
    CP_COLOR = CV_RGB(255, 0, 0);

    kt = NULL;
    am = NULL;
    origImg = NULL;
    displayImg = NULL;
}

/** Destructor */
StateSetter::~StateSetter() {
    cvReleaseImage(&origImg);
    cvReleaseImage(&displayImg);
}

/** 
    Returns a pointer to the instance of this class.
    Don't call the constructor to get an instance of
    this class; instead call this static method.
    @return pointer to the single instance of this class
 */
StateSetter *StateSetter::getInstance() {
    if (theStateSetter == NULL) {
        // First call; create instance
        theStateSetter = new StateSetter();
    }
    return theStateSetter;
}

/**
    Set the kinematic model, appearance model, and background
    image to use.
    @param kt kinematic model
    @param am appearance model
    @param img background image
    @param frame index of frame being initialized
 */
void StateSetter::initialize(KinematicTree *kt, AppearanceModel *am,
                             IplImage *img, int frame) {
    this->kt = kt;
    this->am = am;
    this->frame = frame;
    
    origImg = ImgFun::newImage(cvSize(img->width, img->height), img->depth,
                               img->nChannels);
    cvCopy(img, origImg);

    displayImg = ImgFun::newImage(cvSize(img->width, img->height));

    dragging = false;

    display();
}

/**
    Shows the output in an OpenCV HighGUI window.  The output consists of
    the background image with the appearance model superimposed to
    show the current state.
 */
void StateSetter::display() {
    cvCopy(origImg, displayImg);
    am->draw(displayImg, kt, MODEL_RED, MODEL_GREEN, MODEL_BLUE);

    controlPts.erase(controlPts.begin(), controlPts.end());
    kt->getControlPts(controlPts);

    vector<CvPoint>::iterator itr = controlPts.begin();
    for (; itr != controlPts.end(); itr++) {
        cvCircle(displayImg, (*itr), CP_RADIUS, CP_COLOR, CV_FILLED);
    }

    cvNamedWindow(SS_WINDOW_NAME, CV_WINDOW_AUTOSIZE);
    cvSetMouseCallback(SS_WINDOW_NAME, SSMouseCallback, 0);
    cvShowImage(SS_WINDOW_NAME, displayImg);
}

/**
    Hides the output window. 
 */
void StateSetter::hide() {
    cvDestroyWindow(SS_WINDOW_NAME);
}

/**
    Handles mouse events from the HighGUI display window.  This function
    should not be called directly.  It is called by
    the global function SSMouseCallback(), which is the mouse
    handler for the display window.
    @param event identifier of event type.  This method is only interested
        in left button down and up and mouse move events.
    @param x coordinate of mouse event
    @param y coordinate of mouse event
    @param flags event flags
 */
void StateSetter::mouseHandler(int event, int x, int y, int flags) {
	CvPoint clickPoint = cvPoint(x, displayImg->height - y);
    int i;
    vector<CvPoint>::iterator itr;

	switch(event) {
    case CV_EVENT_LBUTTONDOWN:
		itr = controlPts.begin();
		for (i = 0; itr != controlPts.end(); itr++, i++) {
            if (MathFun::dist((*itr), clickPoint) <= CP_RADIUS) {
                draggedIdx = i;
                dragging = true;
                break;
            }
        }
        break;
    
    case CV_EVENT_LBUTTONUP:
    case CV_EVENT_MOUSEMOVE:
        if (dragging) {
            kt->setStateFromControlPts(clickPoint, draggedIdx);
            display();
        }
        if (event == CV_EVENT_LBUTTONUP) {
            dragging = false;
        }
        break;

    default:
        break;
    }
}

/** @addtogroup bridge_group */
/* @{ */
/**
    Mouse callback for the StateSetter class.  All this does is call
    the method StateSetter::mouseHandler().
    @see StateSetter::mouseHandler()
    @param event identifier of event type.  This method is only interested
        in left button down and up and mouse move events.
    @param x coordinate of mouse event
    @param y coordinate of mouse event
    @param flags event flags
 */
void SSMouseCallback(int event, int x, int y, int flags, void* param) {
    StateSetter::getInstance()->mouseHandler(event, x, y, flags);
}
/* @} */