#ifndef STATE_SETTER_INCLUDED
#define STATE_SETTER_INCLUDED

#include "cv.h"

#include "AppearanceModel.h"
#include "KinematicTree.h"

#define SS_WINDOW_NAME ("ss_window")

/** @addtogroup bridge_group */
/* @{ */
/**
    Manages a KinematicTree, AppearanceModel, and an OpenCV
    HighGui window to allow the user to set the model
    state.  

    This is a singleton class.  There should only ever be
    one instance oft his class.  To access the instance,
    call the getInstance() method to get a pointer to it.
    Don't try calling the constructor (that won't work anyway).
    It is singleton because the HighGUI mouse handlers have
    to be able to call into it.
 */
class StateSetter {
private:
    /** Pointer to the single instance of this class */
    static StateSetter *theStateSetter;

    /** The kinematic model we're setting the state in */
    KinematicTree *kt;

    /** The appearance model */
    AppearanceModel *am;

    /** The video frame we are setting the state for */
    IplImage *origImg;
    
    /** The image displayed in the window; background is
        origImg, foreground is the appearance model */
    IplImage *displayImg;

    /** vector of control points used to manipulate the
        kinematic model state */
    vector<CvPoint> controlPts;

    /** True if one of the control points is being dragged */
    bool dragging;

    /** Index of the control point currently being dragged */
    int draggedIdx;
    
    /** Color used to draw control points */
    CvScalar CP_COLOR;

    /** Radius of the control points */
    enum {CP_RADIUS = 5};   

    /** Color used to draw appearance model */
    enum {MODEL_RED = 255, MODEL_GREEN = 255, MODEL_BLUE = 255};

    StateSetter();
    StateSetter(const StateSetter &ss);
    StateSetter & operator=(const StateSetter &ss);
    ~StateSetter();
    void display();

public:
    /** Index of frame for which state is being initialized */
    int frame;

    static StateSetter * getInstance();
    void initialize(KinematicTree *kt, AppearanceModel *am, IplImage *img,
                    int frame);
    void hide();
    void mouseHandler(int event, int x, int y, int flags);
};
/* @} */


// Mouse callback function
void SSMouseCallback(int event, int x, int y, int flags, void* param);

#endif // !def STATE_SETTER_INCLUDED
